import pandas as pd
import os
from datetime import date

AbrDict = {'Ala':'ala',
        'Arg':'arg',
        'Asp':'asp',
        'Cys':'cys',
        'Glu':'glu',
        'Gly':'gly',
        'His':'his3',
        'Iso':'ile',
        'Leu':'leu',
        'Lys':'lys',
        'Met':'met',
        'Phe':'phe',
        'Pro':'pro',
        'Ser':'ser',
        'Thr':'thr',
        'Try':'trp',
        'Tyr':'tyr',
        'Val':'val'}
# Automatic generated code with Xcal input:
# def Xcal2iMS2(dataframe, Xcalibur, iMS2):
#     """Convert Xcalibur data to iMS2 format.

#     Parameters
#     ----------
#     dataframe : pandas.DataFrame
#         Dataframe containing Xcalibur data.
#     Xcalibur : str
#         Path to Xcalibur file.
#     iMS2 : str
#         Path to iMS2 file.

#     Returns
#     -------
#     None
#         Writes iMS2 file.
#     """
#     # Read Xcalibur file
#     xcal = pd.read_csv(Xcalibur, sep='\t', header=0)
#     # Get Xcalibur data
#     xcal = xcal[['Scan', 'Precursor', 'Precursor Charge', 'Precursor m/z', 'Precursor Intensity']]
#     # Rename columns
#     xcal.columns = ['Scan', 'Precursor', 'Charge', 'm/z', 'Intensity']
#     # Merge dataframes
#     df = pd.merge(dataframe, xcal, on='Scan')
#     # Write iMS2 file
#     df.to_csv(iMS2, sep='\t', index=False)



def Xcal2iMS(XcalFile, Experiment=None, AAIgnore=None):
    '''Convert Xcalibur data to iMS2Flux input format.
    Parameters
    ----------
    XcalFile : str
        Path to Xcalibur file.
        This file is generated by the Xcalibur software and contains one sheet for each fragment, thus with hundreds of sheets.
        Each sheet contains the data for one fragment, e.g. 'Alanine_232'.
    Experiment : list, optional
        Name of the experiment to be used. The default is None, which corresponds to all experiments.
    AAIgnore : list, optional
        List of amino acids to be ignored. The default is None, which corresponds to no amino acids to be ignored.

    Returns
    -------
    outputfile_name : str
        Writes iMS2Flux tab-separated tsv-file.        
    '''
    Today = date.today().strftime('%y%m%d')
    # Extracting path and filename from XcalFile
    head, tail = os.path.split(XcalFile)

    xls = pd.ExcelFile(XcalFile)
    
    # Reading the first sheet to get the experiment identifiers
    myExp = pd.read_excel(xls, header=4, usecols='A')
    myDat = pd.DataFrame()
    for sheet in xls.sheet_names[:-1]:
        myDat[sheet] = pd.read_excel(xls, sheet, header=4, usecols='O', na_values=['NF'])
    
    # Select defined experiments
    if Experiment is not None:
        myExpList = myExp.iloc[:,0].isin(Experiment)
        myExp = myExp.loc[myExpList,:]
        myDat = myDat.loc[myExpList,:]


    # find columns that contain real data
    UseCols = myDat.notna().any()
    # find rows that contain real data
    UseRows = myDat.iloc[:,0].notna()
    myDat = myDat.loc[:, UseCols] 
    myDat = myDat.loc[UseRows,:]
    # get the fragment names
    myFrags_list = myDat.columns
    myDat = myDat.T
    myDat.reset_index(inplace=True, drop=True)
    # Delete rows that contain NaN in experiment identifiers
    myExp = myExp.loc[UseRows,:]
    myDat.columns = myExp.iloc[:,0]

    # Create new dataframe with fragment names and mass, using AbrDict from above...
    AAId = [AbrDict[Frag[:3]] for Frag in myFrags_list]
    Mass = [Frag[-3:] for Frag in myFrags_list]
    mynew = pd.DataFrame([AAId, Mass]).T
    myRes = pd.concat([mynew, myDat], axis=1)

    # Remove defined amino acid fragments
    if AAIgnore is not None:
        myResList = ~myRes.iloc[:,0].isin(AAIgnore)
        myRes = myRes.loc[myResList,:]
    myRes.reset_index(inplace=True, drop=True)


    # keep only the first name-occurence of each fragment
    repeat = myRes[0].drop_duplicates(keep='first')
    myRes[0] = ''
    myRes.iloc[repeat.index,0] = repeat
    myRes.rename(columns={0:'', 1:''}, inplace=True)

    # Write iMS2Flux file
    outputfile_name = os.path.join(head, 'iMS2Flux_' + tail[:-5] + '_' + Today + '.tsv')
    os.remove(outputfile_name) if os.path.exists(outputfile_name) else None
    with open(outputfile_name, 'a') as file:
        file.write('Response Values (Raw Data)')
        file.write('\n')
        myRes.to_csv(file, sep = '\t', index = False)
    return outputfile_name


import numpy as np
import pandas as pd
# import os
# import datetime
from iambcodes.rates import *

def GrowthCheck(myDat, myThresh):
    '''
    Check whether the OD surpases a threshold defined as a multiple from initial OD
    '''
    MaxIdx = np.argmax(myDat)
    Test = True if myDat[MaxIdx]/myDat[0] > myThresh else False
    return Test
############################
#
############################
def GrowthPhase(myDat, myThresh=.05):
    '''
    Selecting the biomass vector were growth was exponential and stationary, but not in death phase.
    '''
    MaxIdx = np.argmax(myDat)
    MyCut = (1-myThresh)*myDat[MaxIdx]
    DeathPhase = myDat[MaxIdx:]<MyCut
    DPIdx = np.where(DeathPhase)[0]
    KeepIdx = MaxIdx+DPIdx[0]-1 if DPIdx.size>0 else len(myDat)
    return myDat[:KeepIdx], myDat.index[:KeepIdx]
############################
#
############################
def GP_PlateData(FileName, GrowthThresh=3, GVcorrect={'GVexp':1, 'eexp':0, 'Background':0}, GPsource='GP'):
    if GPsource == 'GP-de':
        skiprows = 23
        decimal = ','
        TimeUnit = 60
    if GPsource == 'GP-en':
        skiprows = 23
        decimal = '.'
        TimeUnit = 60
    elif GPsource == 'Other':
        skiprows = 0
        decimal = '.'
        TimeUnit = 60
        
    GVexp = GVcorrect['GVexp']
    eexp = GVcorrect['eexp']
    Background = GVcorrect['Background']

    df = pd.read_csv(FileName, skiprows=skiprows , decimal=decimal)
    df = df.drop(labels='Input_Image', axis=1)
    df.dropna(axis=1, inplace=True)
    myTime = pd.Series(df['Time (min)'].values/TimeUnit)
    myGr = df.loc[:, df.columns != 'Time (min)'].apply(lambda x: CorrectedOD(x, GVexp, eexp, Background))
    # removing measurements without correct growth magnitude
    GrowthSelect = myGr.apply(lambda x: GrowthCheck(x, GrowthThresh), axis=0)
    myGr.drop(myGr[myGr.columns[~GrowthSelect.values]], axis=1, inplace=True)

    return myTime, myGr
############################
#
############################
def MinMaxRange(x):
    '''
    Find the indices between minimum and maximum OD.
    
    Arguments:
        x:  numpy vector, OD measurements
        
    Return:
            numpy vector, indices of x
    '''   
    return np.arange(np.argmin(x), np.argmax(x)+1)
############################
#
############################
def MakeBins(x,bins=2):
    '''
    The input vector is binned into equal sized partitions.

    Arguments:
        x:          numpy vector, OD measurements
        bins:       integer, number of bins
    Returns:
        borders:    numpy array, indices of start (column: 0) and end (column: 1) of bins
                    0, if fewer than three measurements would occupy the bins
        
    '''
    mb = int(len(x)/bins)
    if mb > 0:
        mall = np.arange(0,len(x)+1,mb)
        borders = np.vstack([mall[:-1],mall[1:]]).T
    else:
        borders = 0
        
    return borders
############################
#
############################
def SlopeCalc(t, x, bins):
    '''
    Calculation of linear regression on ln(OD) and time.
    
    Arguments:
        t,x:        numpy vector, OD and time measurements
        bins:       numpy array, indices of start (column: 0) and end (column: 1) of bins
        
    Returns:
        (m,c,r2):   m: float, slope; c: float, y-intersection; r2: float, correlation coefficient
        myRange:    numpy vector, updated indices of x used for linear regression
    '''
    myRange = np.arange(bins[0],bins[1])
    y_rate, y_sterr, y_init, y_lim = FitGrowth(t[myRange], x[myRange], law='lin', inlog=True)
    return ((y_rate, y_sterr, y_init, y_lim), myRange)
############################
#
############################    
def DetectR2MaxSingle(t, x, partitions, nSamples=4):
    '''
    Iterative partitioning of the data range to find the bin with the highest correlation coefficient for exponential growth.
    
    Arguments:
        t,x:        numpy vector, OD and time measurements
        partitions: integer, number of bins
        nSamples:   integer, minimum number of samples for each estimation 
    '''
    # Only checking region between min and max OD values
    MaxRange = MinMaxRange(x)
    ttest, xtest = np.array(t[MaxRange]), np.array(x[MaxRange])
    
    # Initiating loop parameters
    counter = 0
    R2_ref = 0
    R2_tst = .0000001
    # Slope_ref = 0
    # Slope_tst = .000001
    myResult = ''
    # Generating new partitions in the data until the correlation coefficient of the new sub-bins is lower than the combined parent bin
    while R2_tst > R2_ref and counter<10:# or R2_tst > .95: 
        # overwriting reference with current solution
        # Slope_ref = Slope_tst
        R2_ref = R2_tst    
        # partitioning new range
        bins = MakeBins(xtest,partitions)
        # Return False if too few data points remain for sensible binning (<4) 
        if bins[0][1]-bins[0][0]>nSamples:
            result = np.array([SlopeCalc(ttest, xtest, mybin) for mybin in bins], dtype=object)
            mySlopes = [res[0][0] for res in result]
            MaxSlpIdx = np.argmax(mySlopes)
            popt = [result[MaxSlpIdx][0][0], result[MaxSlpIdx][0][2]]
            SStdErr = result[MaxSlpIdx][0][1]
            r2, _ = FitR2(ttest[result[MaxSlpIdx][1]], xtest[result[MaxSlpIdx][1]], popt, law='lin', inlog=True)
            R2_tst = r2
            # Storing best solution
            if R2_tst > R2_ref:
                myResult = {'R2':r2, 'Slope': popt[0], 'SlopeStdErr': SStdErr, 'ycorrect':popt[1], 'time':ttest[result[MaxSlpIdx][1]], 'OD':xtest[result[MaxSlpIdx][1]]}

            xtest = xtest[result[MaxSlpIdx][1]]
            ttest = ttest[result[MaxSlpIdx][1]]
            counter += 1
    if not myResult:
        print('Slope estimation failed, try higher growth threshold.')
        
    return myResult
############################
#
############################
def WellFit(time, biomass, bins, law='log', DeathThresh=.05):
#     mybio, myIdx = GrowthPhase(myGr[well], DeathThresh)
#     mytime = time[myIdx]
    if law=='log':
        # extracting relevant biomass values not in death stage
        res = FitGrowth(time, biomass, law='log')
        if res:
            y_rate, y_sterr, y_init, y_lim = res
            PlotPar = [y_rate, y_init, y_lim]
            r2, _ = FitR2(time, biomass, PlotPar, law='log')
            result = {'R2':r2, 'Slope': y_rate, 'SlopeStdErr': y_sterr, 'ycorrect':y_init, 'y_lim': y_lim, 'time':time, 'OD':biomass, 'Law':'log'}
        else:
            result = DetectR2MaxSingle(time, biomass, bins)
            result['Law'] = 'lin'
    else: 
#         print('log failure, using exponential')
        result = DetectR2MaxSingle(time, biomass, bins)
        result['Law'] = 'lin'
#         popt = [result['Slope'], result['ycorrect']]
    return result
############################
#
############################
def CorrectedOD(GV,expo1,expo2,Background):
    '''
        Correction of the OD
    '''
    od = ((GV-Background) ** expo1) * (np.exp(expo2))
    return od
############################
#
############################
def weighted_avg_and_std(values, weights):
    """
    Return the weighted average and standard deviation.

    values, weights -- Numpy ndarrays with the same shape.
    https://stackoverflow.com/questions/2413522/weighted-standard-deviation-in-numpy
    """
    average = np.average(values, axis=1, weights=weights)
    AvTiled = np.tile(average,(values.shape[-1],1)).T
    # Fast and numerically precise:
    variance = np.average((values-AvTiled)**2, axis=1, weights=weights)
    return (average, np.sqrt(variance.astype(np.float64)))


from Bio import SeqIO

AAlist = ['A', 'R', 'N', 'D', 'C', 'Q', 'E', 'G', 'H', 'I', 'L',
          'K', 'M', 'F', 'P', 'S', 'T', 'W', 'Y', 'V']

def extractEC(FName:str) -> dict:
    '''
    Extraction of Gene IDs and EC numbers from input file.
    Parameters
    ----------
    FName : str, 
        e.g. 'EC_Annotation.txt'.

    Returns
    -------
    dict
        Dictionary.

    '''
    # loading reference Usti gene-EC mapping
    # FName = 'EC_Annotation.txt'
    ECList = list()
    GeneID = list()
    ECID = list()
    with open(FName) as file:
        for line in file:
            if '[EC:' in line:
                GeneID.append(line.split('\t')[0]) # the first 10 characters represent the gene ID
                ECID.append(line[line.find('[EC:'):line.rfind(']')+1])
    
    EC_dict = dict(zip(GeneID, ECID))
    return EC_dict
# =============================================================================
# 
# =============================================================================
def processFasta(fastafile:str, EC_dict:dict, TestList:str=['unknown', 'hypothetical']) -> dict:
    # protein descriptions that flag ignorable entries
    # TestList = ['unknown', 'hypothetical']
    # "pangenome_AA_Umay.faa"
    
    # Initializing list in order to use append in loop
    recids = list()
    records = list()
    doubles = list()
    with open(fastafile) as handle:
        for record in SeqIO.parse(handle, "fasta"):
            # ignoring duplicate id records
            if record.id not in recids:
                # ignoring hypothetical, unknown entries
                if not any([test in record.description for test in TestList]):
                    if '[EC:' in record.description:
                        record.annotation = record.description[record.description.find('[EC:'):record.description.rfind(']')+1]
                    elif record.id in EC_dict.keys():
                        record.annotation = EC_dict[record.id]
                    records.append(record)
                    recids.append(record.id)
            else:
                doubles.append(record.id)
    
    # storing the entries in the fasta with non-unique id
    DoubleFile = 'Pan_Double_IDs.txt'
    with open(DoubleFile, 'w') as outfile:
        outfile.write('\n'.join(doubles))
        
    rec_dict = SeqIO.to_dict(records)
    return rec_dict
# =============================================================================
# 
# =============================================================================
def RelAASeq(sequence:list):
    myRelAA = [sequence.count(AAid)/len(sequence) for AAid in AAlist]
#     pd.Series(sequence).value_counts()/len(sequence)   
    return myRelAA
# =============================================================================
# 
# =============================================================================
def RelAA4RecDict(rec_dict:dict):
    
    # import pandas as pd
    import numpy as np
    
    SumRelAA = np.zeros(len(AAlist))
    for record in rec_dict.items():
        myRelAA = RelAASeq(list(record[1].seq))
        SumRelAA += myRelAA
    return SumRelAA / len(rec_dict)


import pandas as pd
import numpy as np
import itertools
import string

import matplotlib.pyplot as plt
from matplotlib import colors

def ImportXls(FName:str):
    ftmp = pd.ExcelFile(FName)
    # extracting metadata information on number of replicates and measurements
    MetaDat = pd.read_excel(FName, ftmp.sheet_names[0])
    # Replicate number
    myReps = MetaDat.loc[MetaDat[MetaDat.columns[0]]=='Replicates:'][MetaDat.columns[1]].values.astype('int')
    # measurement points
    myMeas = MetaDat.loc[MetaDat[MetaDat.columns[0]]=='Measurements:'][MetaDat.columns[1]].values.astype('int')
    
    # extracting data from different sheets
    DataSheets = ftmp.sheet_names[1:]
    BiologAll = np.zeros([len(DataSheets),96])
    TimeVec = np.zeros([len(DataSheets)])
    for indx, DataSheet in enumerate(DataSheets):
        df = pd.read_excel(FName, sheet_name=DataSheet, header=None)
        Dtmp_arr = df.iloc[2:11,1:13].values
        Dtmp_vec = np.ndarray.flatten(Dtmp_arr)
        BiologAll[indx,:] = Dtmp_vec
        TimeVec[indx] = df.iloc[0,1]

    # calculating mean and std for replicates over all measurements
    BiologMean = np.zeros([int(myMeas), 96])
    BiologStdv = np.zeros([int(myMeas), 96])
    for Meas in range(int(myMeas)):
        Tindx = np.arange(Meas, Meas+myReps)
        BiologMean[Meas,:] = np.nanmean(BiologAll[Tindx,:], axis=0)
        BiologStdv[Meas,:] = np.nanstd(BiologAll[Tindx,:], axis=0)

    # # fancy rearrangement of columns to represent mean+std of each replicate
    # # https://stackoverflow.com/questions/20265229/rearrange-columns-of-numpy-2d-array
    Tvec = np.unique(TimeVec).reshape([-1,1])
    # Alpha = [let for let in string.ascii_uppercase[:8]]
    # Num = list(map(str,range(1,13)))
    Columns = WellIDs()
    Columns.insert(0,'Time_h')
    BiologMean_df = pd.DataFrame(data=np.hstack([Tvec,BiologMean]), columns=Columns)
    BiologStdv_df = pd.DataFrame(data=np.hstack([Tvec,BiologStdv]), columns=Columns)
    
    return BiologMean_df, BiologStdv_df
 

def VisualizeWell(myArr, plot_title=None, plot_z_name=None, export_file=None, cmap='viridis'):
    WellShape = np.reshape(myArr, (8,12))

    if cmap == 'discrete5':
#    https://stackoverflow.com/questions/9707676/defining-a-discrete-colormap-for-imshow-in-matplotlib
    	cmap = colors.ListedColormap(['orange', 'yellow', 'blue', 'green'])
#     	bounds = np.arange(0,5)

    plt.figure()
    im = plt.imshow(WellShape, interpolation='none', aspect='equal', cmap=cmap)
    plt.title(plot_title)
    plt.colorbar(label=plot_z_name)
    
    ax = plt.gca();
    
    # Major ticks
    ax.set_xticks(np.arange(0, 12, 1))
    ax.set_yticks(np.arange(0, 8, 1))
    
    # Labels for major ticks
    ax.set_xticklabels(np.arange(1, 13, 1))
    ylabel = [let for let in string.ascii_uppercase[:8]]
    ax.set_yticklabels(ylabel)
    
    # Minor ticks
    ax.set_xticks(np.arange(-.5, 12, 1), minor=True)
    ax.set_yticks(np.arange(-.5, 8, 1), minor=True)
    
    # Gridlines based on minor ticks
    ax.grid(which='minor', color='k', linestyle='-', linewidth=2)
    if export_file:
        plt.savefig(export_file)
    # # visualizing
    # x = np.arange(1,9)
    # y = [let for let in string.ascii_uppercase[:8]]
    # plt.yticks(x, y)
    # plt.imshow(WellShape, extent= [1, 12, 8, 1])
    plt.show()
    
def VisualizeWellSubstrate(myArr, PMIDs, PMTruth, plot_title=None, plot_z_name=None, export_file=None):
    WellShape = np.reshape(myArr, (8,12))

    plt.figure()
    im = plt.imshow(WellShape, interpolation='none', aspect='equal')
    plt.title(plot_title)
    plt.colorbar(label=plot_z_name)
    
    ax = plt.gca();
    
    # Major ticks
    ax.set_xticks(np.arange(0, 12, 1))
    ax.set_yticks(np.arange(0, 8, 1))
    
    # Labels for major ticks
    ax.set_xticklabels(np.arange(1, 13, 1))
    ylabel = [let for let in string.ascii_uppercase[:8]]
    ax.set_yticklabels(ylabel)
    
    # Minor ticks
    ax.set_xticks(np.arange(-.5, 12, 1), minor=True)
    ax.set_yticks(np.arange(-.5, 8, 1), minor=True)
    
    # Gridlines based on minor ticks
    ax.grid(which='minor', color='k', linestyle='-', linewidth=2)
    
    # loop over all wells to label them with substrate ids
    for indx, mypos in enumerate([[r[0],r[1]] for r in itertools.product(np.arange(8), np.arange(12))]):
        if PMTruth[indx]:
            plt.text(mypos[1], mypos[0], PMIDs[indx], va='center', ha='center', fontsize=8)
            
    if export_file:
        plt.savefig(export_file)
    plt.show()
    
def VisualizeConfMatrix(myList, export_file=None):
    OrdCount = list()
    for ordinal in range(4,0,-1):
        OrdCount.append(sum(myList==ordinal))
        
    OrdCount_rel = OrdCount/sum(OrdCount)
	# in PM_df the order of properties is:
	# 0: metabolite not in model
	# 1: Biolog-, iMod+
	# 2: Biolog+, iMod-
	# 3: Biolog-, iMod-
	# 4: Biolog+, iMod+
	# To visualize this as a confusion matrix, the elements 1-4 are normalized to their sum and rearranged as matrix with x-axis representing +,- for Biolog and y-axis +,- for iMod
    OrdCount_c = np.array([OrdCount[ConfIdx] for ConfIdx in [0,3,2,1]])
    plt.imshow(OrdCount_c.reshape(2,2))
    plt.xticks([0,1], ['True','False'])
    plt.xlabel('BIOLOG')
    plt.yticks([0,1], ['True','False'])
    plt.ylabel('iModel')
    for myNum in zip([[r[0],r[1]] for r in itertools.product(np.arange(2), np.arange(2))], OrdCount_c):
        print(myNum)
        plt.text(myNum[0][1], myNum[0][0], myNum[1], va='center', ha='center', fontsize=8)
    plt.title('Growth comparison for {} of {} substrates'.format(sum(OrdCount_c), len(myList)))
    if export_file:
        plt.savefig(export_file)
    plt.show()

 
def SlopeCalc(myArr, zero_max=False):
    '''
    Calculation of linear slopes. First column must be time. If desired ('zero_max' True) only the difference from zero to the last measurement is calculated.
    '''
    myTimes = myArr[:,0]
    myDat = np.delete(myArr,0,1)

    if zero_max:
        mySlope = myDat[-1,:]/myTimes[-1]
    else:
        mySlope = (myDat[1:,:] - myDat[:-1,:]) / np.tile((myTimes[1:]-myTimes[:-1]).reshape(-1,1),(1,myDat.shape[1]))
    
    return mySlope

def WellIDs():
    '''Generates the Well ids'''
    Alpha = [let for let in string.ascii_uppercase[:8]]
    Num = list(map(str,range(1,13)))
    return [r[0]+r[1] for r in itertools.product(Alpha, Num)]

def ConvertMatrixList(myList:list, dimension=[12,8]):
    '''
    Converting a 1D list which was generated from a matrix with stacked columns into a 1D list with stacked rows.

    Parameters
    ----------
    myList : list
        1D representation by stacked columns of matrix.
    dimension: list
        Dimension (columns, rows) of the original matrix. By default a 96-well plate with 12x8
    
    Returns
    -------
        1D representation of stacked rows.

    '''
    myList_matrix = np.array(myList).reshape(dimension)
    return list(np.transpose(myList_matrix).flatten())



import os
from datetime import date

def Create_New_ConfFile(attributes = list, workflow = str, target_dir = str):
    #Aufgabe der Funktion ist es, die liste anzunehmen und eine .txt datei zu erstellen
    
    """
    Parameters:
    
    attributes = list of strings that end up as keys when loading the config file
    
    workflow = string of the type of worklow the conffile is generated for
    
    target_dir = string of the target directory
    
    Today = Date of use. Needs datetime-module
    
    """
    
    #generate the date
    Today = date.today().strftime('%y%m%d')
    
    
    #Check if workflow is correct
        #function Check_Workflow_Type has a list of all available Workflow types and checks if the chosen worflow
        #is part of the list
        #if true, continue
        #if false, error message about a false worklow
    if Check_Workflow_Type(workflow) == False:
        print("WARNING! The chosen worklow does not exist.")
        return False
    
    #Check if attributes is a list
        #if true, continue
        #if not, error message about wrong type of attribute
    if type(attributes) == list:
        #check if list contains only strings
        #if true continue
        #if false, error message about wrong data types
        for i in attributes:
            if type(i) != str:
                print("WARNING! Incorrect data-type found. All Parameters for the config-file have to be in string-format!")
                return False
    else:
        print("WARNING! Function recieved incorrect information. Please check if everything is in order!")
        return False
    
    #write and open new file in directory indicated by Workflow
        #alternative: addition string input with the target directory
    #take list and writerow with one position per row
    
    ConfComment = "#This is a freshly generated configuration file for a {} workflow.\n#Please fill in the parameters according to the {} notebook.".format(workflow, workflow)
    
    ConfName = '{}_JUDAS_{}_{}_config.txt'.format(Today, workflow, "GENERATED")
    ConfAddress = os.path.join(target_dir, ConfName)
    print(ConfAddress)
    with open(ConfAddress, 'w') as f:
        print(ConfComment, file=f)
        for key in attributes:
            print('{}:'.format(key), file=f)
    #done
    
    
    return True


"""
Function to check if a workflow is legitimate or not

Sends back TRUE if workflow is known
Sends back FALSE is worklow is not recognized
"""
def Check_Workflow_Type(workflow = str):
    """
    Parameters:
    listofworklows: list of all available workflows
    """
    listofworkflows = ["BiologData",
                       "DoE",
                       "GrowthProfiler",
                       "MFA",
                       "RatesYields"
                      ]
    
    if workflow in listofworkflows:
        return True
    else:
        return False



import os
from datetime import date
import pandas as pd

def Create_New_ConfFile(workflow = str, target_dir = str):
    #task of the function is to generate a .txt file for the requested workflow
    #An excel file servng as a template library needs to be in the main-folder of iambjudas
    
    """
    Parameters:
    
    ##attributes = list of strings that end up as keys when loading the config file
        Comment: removed. Attributes are now stored as a library of excel sheets
    
    workflow = string of the type of worklow the conffile is generated for
    
    target_dir = string of the target directory
    
    Today = Date of use. Needs datetime-module
    
    """
    
    #generate the date
    Today = date.today().strftime('%y%m%d')
    
    
    #Check if workflow is correct
        #function Check_Workflow_Type has a list of all available Workflow types and checks if the chosen worflow
        #is part of the list
        #if true, continue
        #if false, error message about a false worklow
    if Check_Workflow_Type(workflow) == False:
        print("WARNING! The chosen worklow does not exist.")
        return False
    
    #
    #load of library.xlsx with Prewritten Files
    try:
        file = "220718_configfile_library.xlsx"
        df = pd.read_excel(file, sheet_name=workflow, keep_default_na=False)
    except FileNotFoundError:
        print("ERROR!\nThe config-file library was not found.\n\n")
        return False
    
    #write and open new file in directory indicated by Workflow
        #alternative: addition string input with the target directory
    #take list and writerow with one position per row
     
    ConfName = '{}_JUDAS_{}_{}_config.txt'.format(Today, workflow, "GENERATED")
    ConfAddress = os.path.join(target_dir, ConfName)
    print("ATTENTION!\n\n Your new config-file was generated in the folder "+ConfAddress)
    with open(ConfAddress, 'w') as f:
        for key in df.Parameters:
            print('{}'.format(key), file=f)
    #done
    
    
    return True


"""
Function to check if a workflow is legitimate or not

Sends back TRUE if workflow is known
Sends back FALSE is worklow is not recognized
"""
def Check_Workflow_Type(workflow = str):
    """
    Parameters:
    listofworklows: list of all available workflows
    """
    listofworkflows = ["BiologData",
                       "DoE",
                       'GSMM_Quality_Control',
                       "GrowthProfiler",
                       "MFA",
                       "RatesYields"
                      ]
    
    if workflow in listofworkflows:
        return True
    else:
        return False


import numpy as np
import csv
import re
from cobra.io import read_sbml_model
from cobra import Reaction, Metabolite

###############################################################################
###############################################################################
###############################################################################
#
# Basic functions working on strings without cobra model
#
###############################################################################
###############################################################################
###############################################################################

def countCinFormula(formula):
    '''
    Counting C-atoms of metabolite in cobra model
    
    Argument:
    formula: string, the metabolite formula from the cobra model
    
    Return:
    int, number of C-atoms in formula
    '''
    return int(re.search(r'C(.*)H',formula).group(1)) if re.search(r'C(.*)H',formula).group(1) else int(1)

###############################################################################
###############################################################################
###############################################################################
#
# Functions working on cobra model objects but without the explicit model
#
###############################################################################
###############################################################################
###############################################################################
def MetNewFromExist(MetIn, CompSep):
    '''
    Generate new cobra external metabolite object from input cobra metabolite.

    Parameters
    ----------
    MetIn : cobra metabolite object
        Internal metabolite for which an external metabolite is generated.
    CompSep: string='_' or '[]', separator of metabolite and compartment in id

    Returns
    -------
    Metout : cobra metabolite object
        External metabolite object.

    '''
    Metout = Metabolite(f'{MetIn.id.split(CompSep[0])[0]}_e' if len(CompSep) == 1 else f'{MetIn.id.split(CompSep[0])[0]}[e]',
                       formula = MetIn.formula,
                       name = MetIn.name,
                       compartment = 'e',
                       charge = MetIn.charge)
    Metout.annotation = MetIn.annotation
    return Metout

def CreateTransReact(MetIn, CompSep='_'):
    '''
    Create transfer reaction from internal to external metabolite.

    Parameters
    ----------
    MetIn : cobra metabolite object
        Internal cobra metabolite object.
    CompSep: string='_'(default) or '[]', separator of metabolite and compartment in id

    Returns
    -------
    reaction : cobra reaction object
        Transport reaction from cytoplasm to extracellular.

    '''
    reaction = Reaction('Trans_{}'.format(MetIn.id),
                       name = 'Transport c<->e {}'.format(MetIn.name),
                       lower_bound = -1000,
                       upper_bound = 1000)
    MetOut = MetNewFromExist(MetIn, CompSep)
    reaction.add_metabolites({MetIn:-1.0, MetOut:1.0})
    
    return reaction
###############################################################################
###############################################################################
###############################################################################
#
# Functions working on cobra model objects
#
###############################################################################
###############################################################################
###############################################################################

def set_ObjGSMM(model, Product):
    '''
    Adding boundary flux for a desired product and setting the objective reaction to the product in the model.
    
    Arguments:
    model: cobra model, iUma22
    Product: string, model metabolite ID
    
    Return:
    cobra model, updated with a demand flow for the product and optimization of the new demand reaction.
    '''
    model.add_boundary(model.metabolites.get_by_id(Product), type='demand')
    model.objective = 'DM_{}'.format(Product)
    return model

def set_SubFlux(model, SubOn, SubOff, Flux):
    '''
    Setting the desired substrate in the model.
    
    Arguments:
    model: cobra model, iOpol909
    SubOn: string/list, model metabolite ID
    SubOff:string, metabolite to be set to zero
    Flux: float, desired substrate uptake rate
    
    Return:
    cobra model, updated with designed (co-) substrate uptake rates.
    '''
    model.reactions.get_by_id(SubOff).lower_bound = 0
    model.reactions.get_by_id('Ex_{}'.format(SubOn[0])).lower_bound = -Flux
    return model

def TestSubstrate(model, EX_Sub_Act:list, EX_Sub_Off=['EX_glc__D_e'], Rate=10):
    '''
    Simulation of growth rate with defined substrates for growth and substrates not used for growth.

    Parameters
    ----------
    model : cobra model
        Genome scale metabolic model.
    EX_Sub_Act : list of cobra exchange reaction
        List of the substrate exchange reactions to be tested, typically ['EX_glc__D_e'].
    EX_Sub_Off : list of cobra extracellular metabolite names, optional
        DESCRIPTION. The default is ['EX_glc__D'].
    Rate : TYPE, optional
        DESCRIPTION. The default is 10.

    Returns
    -------
    TYPE
        DESCRIPTION.

    '''

    with model:
        medium = model.medium
        # deactivating uptake from off substrates
        medium.update({mySub: 0.0 for mySub in EX_Sub_Off})
        # setting new substrate uptake rates
        medium.update({mySub: Rate for mySub in EX_Sub_Act})
        model.medium = medium
        return round(model.slim_optimize(),2)


def MetSharedRcts(model, Met1, Met2):
    '''
    This function lists all reactions which are shared by two metabolites
    
    Parameters
    ----------
    model : cobra model
        The cobra model for analysis.
    Met1 : str
        Metabolite id.
    Met2 : str
        Metabolite id.

    Returns
    -------
    list, List of shared reactions.

    '''
    return [reaction for reaction in model.metabolites.get_by_id(Met1).reactions if reaction in model.metabolites.get_by_id(Met2).reactions]

def MetFluxConnect(model, Met1, Met2, SubOff='EX_glc__D_e', ngam=False, slim=True):
    '''
    Test whether there exists a flux between two metabolites. The first metabolite is set as source and FBA is optimized to the second metabolite.

    Parameters
    ----------
    model : cobra model
        .
        The cobra model for analysis.
    Met1 : str
        Metabolite id.
    Met2 : str
        Metabolite id.
    SubOff: str
        Substrate id to be set to zero
    ngam: str
        non growth associated maintenance to be set to zero, because checking close metabolites might not fullfill the base ATP demand in ngam
    slim : Boolean
        Decision whether only the value of the objective is reported via slim_optimize, or the full flux-reaction list.

    Returns
    -------
    list, List of reactions and flux values.

    '''
    with model:
        # medium = dict.fromkeys(model.medium, 0)
        medium = model.medium
        medium[SubOff] = 0
        model.medium = medium
        model.add_boundary(model.metabolites.get_by_id(Met1), type='sink')        
        model.add_boundary(model.metabolites.get_by_id(Met2), type='sink')
        if ngam:
            model.reactions.get_by_id(ngam).lower_bound = 0
        model.objective = 'SK_{}'.format(Met2)
        if slim:
            flux = model.slim_optimize()
        else:
            flux = model.optimize()
            
        return flux

###############################################################################
###############################################################################
###############################################################################
#
# BIOLOG related functions
#
###############################################################################
###############################################################################
###############################################################################

def CSVexport(myDict, FName='BiologGrowthTest'):
    fields = ['id','name','id_e','id_c','growth','CL_Growth']
    with open(FName, 'w') as f:
        w = csv.DictWriter(f, fields)
        w.writeheader()
        for k in myDict:
            w.writerow({field: myDict[k].get(field) or k for field in fields})

def CheckPMinGSMM(model, PM_df, EX_Sub_Off=['EX_glc__D_e']):
    Substrate_list = list()
    ResultID = list()
    
    for Indx,Sub in enumerate(PM_df['BiGG'].astype(str)):
        Sub_e = '{}_e'.format(Sub)
        # Sub_c = '{}_c'.format(Sub)
        Sub_eIdx = np.where([Sub_e==met.id for met in model.metabolites])[0]
        # Sub_cIdx = np.where([Sub_c==met.id for met in model.metabolites])[0]
        # first testing whether growth was detected in biolog or not
        # if growth was detected, we give our best to check substrate metabolization, otherwise not
        # if PM_df['Growth'][Indx]:
            # testing whether biolog metabolite is extracellular or cytoplasmic
        if Sub_eIdx.size>0: # and Sub_cIdx.size>0 
            # metabolite is extracellular, 
            # existence of exchange reaction unclear
            EX_SubRct = 'EX_{}'.format(Sub_e)
            try: 
                myGrowth = TestSubstrate(model, [EX_SubRct], EX_Sub_Off)    
            except KeyError:
                myGrowth = 0
#                 model.reactions.get_by_id(EX_SubRct)
#             except KeyError:
# #                with model:
#                 model.add_boundary(model.metabolites.get_by_id(Sub_e), type='exchange', lb=0, ub=1000)
#                 myGrowth = TestSubstrate(model, [EX_SubRct], EX_Sub_Off)
#                     # Substrate_list.append(myGrowth)
#     #                 print('New exchange reaction for {} with growth: {}'.format(Substrate, myGrowth))        
#             else:
            # myGrowth = TestSubstrate(model, [EX_SubRct], EX_Sub_Off)    
                # Substrate_list.append(myGrowth)

#         elif Sub_cIdx.size>0 and Sub_eIdx.size<1: 
#             # metabolite is cytoplasmic but not extracellular
#             myMet = model.metabolites.get_by_id(Sub_c)
#             myTransport = CreateTransReact(myMet)
#             EX_SubRct = 'EX_{}'.format(Sub_e)
# #            with model:
#             model.add_reactions([myTransport])
#             model.add_boundary(model.metabolites.get_by_id(Sub_e), type='exchange', lb=0, ub=1000)
#             myGrowth = TestSubstrate(model, [EX_SubRct], EX_Sub_Off)
#                 # Substrate_list.append(myGrowth)

        # if Sub_cIdx.size<1 and Sub_eIdx.size>0:
        #     metabolite is only extracellular (unlikely)
        #     Substrate_list.append('Z')
                                       
        # if Sub_cIdx.size<1 and Sub_eIdx.size<1:
        #     # metabolite not in GSMM
        #     Substrate_list.append('Z')

#         else:
#             myGrowth = -1
#     # no growth occured in Biolog:
#     else:
#         if Sub_eIdx.size>0: # and Sub_cIdx.size>0 
#             # metabolite is extracellular, 
#             # existence of exchange reaction unclear
#             EX_SubRct = 'EX_{}'.format(Sub_e)
#             try: 
#                 model.reactions.get_by_id(EX_SubRct)
#             except KeyError:
# #                with model:
#                 model.add_boundary(model.metabolites.get_by_id(Sub_e), type='exchange', lb=0, ub=1000)
#                 myGrowth = TestSubstrate(model, [EX_SubRct], EX_Sub_Off)
#                     # Substrate_list.append(myGrowth)
#     #                 print('New exchange reaction for {} with growth: {}'.format(Substrate, myGrowth))        
#             else:
#                 myGrowth = TestSubstrate(model, [EX_SubRct], EX_Sub_Off)
#                 # Substrate_list.append(myGrowth)
        else:
            myGrowth = 0
            
        Substrate_list.append(myGrowth)
        ResultID.append(CheckWellinGSMM(myGrowth,PM_df.loc[Indx, 'Growth']))
    return Substrate_list, ResultID, model


def CheckWellinGSMM(myGrowth, myBool):
    '''
    Return categorical number how the substrate is used in the model.
    0: not in model, missing in model
    1: bio:-; model:+, false positive
    2: bio:+; model:-, false negative
    3: bio:-; model:-, true negative
    4: bio:+; model:+, true positive

    Parameters
    ----------
    myGrowth : number
        model growth rate on substrates in biolog.
    myBool : boolean
        growth in biolog.

    Returns
    -------
    categorical of correlation Biolog+GSMM.

    '''
    if myBool and myGrowth >= .01:
#         growth occured in well and model
        return 4
    elif not myBool and any([myGrowth < .01, np.isnan(myGrowth)]):
#         no growth in well and no growth in model
        return 3
    elif myBool and any([myGrowth < .01, np.isnan(myGrowth)]):
#       growth occured in well not model
        return 2
    elif not myBool and myGrowth >= .01:
#         no growth in well but growth in model
        return 1
#     elif myGrowth < 0:
# #        no growth in well and model lacks metabolite     
#         return 0




import numpy as np
import matplotlib.pyplot as plt
from pandas.plotting._matplotlib.style import get_standard_colors
from scipy import stats
from scipy.optimize import curve_fit


def plot_multi(data, cols=None, spacing=.1, **kwargs):
# source:
# https://stackoverflow.com/questions/11640243/pandas-plot-multiple-y-axes

    # Get default color style from pandas - can be changed to any other color list
    if cols is None: cols = data.columns
    if len(cols) == 0: return
    colors = get_standard_colors(num_colors=len(cols))

    fig, ax = plt.subplots()
    # First axis
    ax.plot(data.loc[:, cols[0]].values, data.loc[:, cols[1]].values, **kwargs)
#    ax = data.loc[:, cols[0]].plot(label=cols[0], color=colors[0], **kwargs)
    ax.set_ylabel(ylabel=cols[1])
    lines, labels = ax.get_legend_handles_labels()

    for n in range(2, len(cols)):
        # Multiple y-axes
        ax_new = ax.twinx()
        ax_new.spines['right'].set_position(('axes', 1 + spacing * (n - 1)))
        data.loc[:, cols[n]].plot(ax=ax_new, label=cols[n], color=colors[n % len(colors)], **kwargs)
        ax_new.set_ylabel(ylabel=cols[n])
        
        # Proper legend position
        line, label = ax_new.get_legend_handles_labels()
        lines += line
        labels += label
        
    ax.legend(lines, labels, loc=0)
    return ax


def SubVector(StartEnd, Vector):
    '''
    Extract a sub-vector based on start and end values. The output are the indices in the input vector.
    '''
    Above = Vector >= StartEnd[0]
    Below = Vector <= StartEnd[1]
    Select = np.logical_and(Above, Below)
    return np.arange(len(Vector))[Select]
    

def FitGrowth(time, ydat, law='lin', inlog=False, p0=None):
    '''
    Curve fitting of a measurement over time to determine rate and standard deviation. Different laws can be chosen.
    See also https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.curve_fit.html
    
    Arguments:
        time:       np vector, time points
        ydat:       np vector, measurements
        law:        str, 'lin' = linear relation; 'exp' = exponential law; 'log' = logarithmic growth law, Verhulst
        inlog:      Boolean, whether input data should be logarithmized
        p0:         list, initial conditions for the estimator
        
    Returns:
        y_rate:     np vector, rate
        y_sterr:    standard deviation
        y_init:     initial value of y
    
    '''
    
    if law == 'lin':
        if inlog:
            ydat = np.log(ydat)
        y_rate, y_init,_,_, y_sterr = stats.linregress(time, ydat)
        y_lim = None
    elif law == 'exp':
        ExpGrowth = lambda t, r, y0: y0*np.exp(r*t)        
        popt, pcov = curve_fit(ExpGrowth, time, ydat, p0=p0)
        y_rate, y_init = popt[0:2]
        y_sterr = np.sqrt(np.diag(pcov))[0]
        y_lim = None
    elif law == 'log':
        LogGrowth = lambda t, r, y0, K: y0*K / (y0 + (K-y0)*np.exp(-r*t))
        try:
            popt, pcov = curve_fit(LogGrowth, time, ydat, p0=p0)
            y_rate, y_init, y_lim = popt[0:3]
            y_sterr = np.sqrt(np.diag(pcov))[0]
        except RuntimeError:
            return False
    return y_rate, y_sterr, y_init, y_lim

def FitR2(time, ydata, popt, law='lin', inlog=False):
    '''
    Calculated the R2 and sum of squares of a fit.
    https://stackoverflow.com/questions/19189362/getting-the-r-squared-value-using-curve-fit

    Parameters
    ----------
    time : np vector
        time points of the experiment.
    ydata : np vector
        measurements.
    popt : np vector, with rate, y_init, and optional y_lim for log func
        DESCRIPTION.
    law : str, optional
        'lin' = linear relation, default; 'exp' = exponential law; 'log' = logarithmic growth law, Verhulst.
    inlog : boolean, optional
        whether input data should be logarithmized. The default is False.

    Returns
    -------
    R2 : np vector.
        correlation coefficient
    SSR : np vector.
        sum of squared residuals

    '''
    if inlog:
        ydata = np.log(ydata)
    if law == 'lin':
        func = lambda time, a, b: a*time+b 
    elif law == 'exp':
        func = lambda t, r, y0: y0*np.exp(r*t)
    elif law == 'log':
        func = lambda t, r, y0, K: y0*K / (y0 + (K-y0)*np.exp(-r*t))

    
    residuals = ydata- func(time, *popt)
    ss_res = np.sum(residuals**2)
    ss_tot = np.sum((ydata-np.mean(ydata))**2)
    r2 = 1 - (ss_res / ss_tot)
    
    return r2, ss_tot

def PlotFit(time, ydat, par, sterr, xyunits, law='lin', inlog=False, export_file=None, kwargs={}):
    '''
    Curve fitting of a measurement over time to determine rate and standard deviation. Different laws can be chosen.

    Arguments:
        time:       np vector, time points
        ydat:       np vector, measurements
        par:        np vector, with rate, y_init, and optional y_lim for log func
        sterr:      np number, standard error of the rate
        xyunits:    list, x and y-axis units
        law:        str, 'lin' = linear relation; 'exp' = exponential law; 'log' = logarithmic growth law, Verhulst
        inlog:      Boolean, whether input data should be logarithmized
        export_file:str, filename of plot
    
    Returns:
        Figure of rate law with data.

    '''
    if inlog:
        ydat = np.log(ydat)
    plt.plot(time, ydat, 'gx', label='data', **kwargs)
    if law == 'lin':
        func = lambda t, a, b: a*t + b
    elif law == 'exp':
        func = lambda t, r, y0: y0*np.exp(r*t)
    elif law == 'log':
        func = lambda t, r, y0, K: y0*K / (y0 + (K-y0)*np.exp(-r*t))
    
    # we sample more points
    ttime = np.linspace(time[0], time[-1], len(time)*10)
    plt.plot(ttime, func(ttime, *par), 'g-', label='fit: rate={:.2f}+/-{:.2f}'.format(par[0], sterr), **kwargs)
    plt.xlabel(xyunits[0])
    plt.ylabel(xyunits[1])
    plt.legend()
    if export_file:
        plt.savefig(export_file)
    # plt.show()

def read_ConfFile(FileAddress:str):
    '''
    Reading JUDAS configuration files. 

    Parameters
    ----------
    FileAddress : str
        Full address of configuration file.

    Returns
    -------
    Par_Dict : dict
        Dictionary with all user defined variables in the respective workflow.

    '''
    Par_Dict = dict()
    with open(FileAddress) as fp:
        mylines = fp.read().splitlines()
        for myline in mylines:
            if not myline.startswith('#'):
                (key, val) = myline.split(':', 1)
                Par_Dict[str(key.strip())] = val.strip()
    return Par_Dict


import setuptools

setuptools.setup(
    name='iambcodes',
    version='0.1.11',
    author='Ulf Liebal',
    author_email='ulf.liebal@rwth-aachen.de',
    description='Functions for data analysis support at the iAMB in RWTH Aachen.',
    keywords='biotechnology',
    url='https://git.rwth-aachen.de/ulf.liebal/iambcodes',
    package_dir={'': 'src'},
    packages=setuptools.find_packages(where='src'),
    python_requires='>=3.6',
    install_requires=[
        'xlrd>=2.0.1',
        'numpy>=1.22.0',
        'cobra>=0.22.1',
        'biopython>=1.78',
        'matplotlib>=3.3.4',
        'scipy>=1.9.0',
    ],
)



import ipywidgets as widgets
import random as rd
import numpy as np
import sympy
import copy
from IPython.display import Image


##########
##########
##########
#class for homework 3
##########
##########
##########

class homework_three():
#class homework_three_test():
    def __init__(self, stdID):
        # Check if Student-ID is an integer and if the function is to be run in "test-mode"
        
        #'TEst-Mode' is activated by the Std-ID input: 123456789
        # For this to be viable, 123456789 CANNEVER be a viable student-id
        if type(stdID) == int and stdID == 123456789:
            self.stdID = "testing"
        elif type(stdID) == int and stdID != 123456789:
            self.stdID = stdID
        else:
            print("Student ID required to be an integer.\nPlease try again!\n\n")
            pass
        
        # How many rngs are to be created for new/changed parameters
        self.numberofrandomnumbers = 6
        
        # Create RNs on the basis of the studentID
        self.initiate_homework()
        
        # Create variables using the generated RNs
        self.create_question_variables()
        
        # List of self.parameters:
        #self.stID
        #self.listofrndms
        
        #self.newYxs
        #self.newH
        #self.newO
        #self.newN
        
        #self.newRO2
        #self.newRCO2
        
        self.hq1tgl = 0 #help_question_1_toggle
        self.hq2tgl = 0 #help_question_2_toggle
        
        # build widgets for the questions
        self.question_one_display = self.build_question_one()
        self.question_two_display = self.build_question_two()
        
        
    def initiate_homework(self):
    # Parameters:
    # stdID is an integer used to seed the rng
        # IF stdID is given the string "test", it's supposed to return the string "test"
        # This string is then used to thest functions using standard values
    # counter is an integer used to define how many rngs are created.
        #print("Matriculation-Number accepted.\n\nGenerating random numbers..")
        #print("Matriculation-Number accepted.\n\nGenerating personalized homework parameters..")
        if self.stdID == "testing":
            print("#####\n\nTesting functionality\n\n#####")
            print("#####\n\The Std-ID is not a valid Student-ID\n and CAN'T generate a valid code!\n\n#####")
            # NOTE: no self.listofrndms is created in test mode
        else:
        #RNG for this homework
        #IMPORTANT: rd.seed() is set GLOBALLY, as far as we observed
        #this means: the solution-codes are ALWAYS IDENTICAL FOR A UNIQUE STid
        #if rd.seed() is not refreshed after generation of listofrndms.
        #In this version, the generated solution-codes are supposed to be random, therefore. rd.seed() gets refreshed
            rd.seed(self.stdID)
            self.listofrndms = [rd.random() for i in range(self.numberofrandomnumbers)] #create new random numbers that can be used to modify parameters
            rd.seed()
        return
    
    def create_question_variables(self):
        #function returns all generated homework parameters and a boolean 
        #thats used to toggls help_question_1_toggle to  0 (helpoption was not used)
        if self.stdID != "testing":
            #print("\nGenerating personalized homework parameters...\n")
            ###PARAMATERS for QUESTION 1
            self.newYxs = 1+round(self.listofrndms[0],1) #range 1.0-2.0
            self.newH = 1.5+round((self.listofrndms[1]*0.5),1) #range 1.5-2.0
            self.newO = 0.3+round((self.listofrndms[2]*0.4),1) #range 0.3-0.7
            self.newN = 0.1+round((self.listofrndms[3]*0.2),1) #range 0.1-0.3

            ###PARAMATERS for QUESTION 2
            self.newRO2 = 3000+round((self.listofrndms[4]*500)) #range 3000-3500)
            self.newRCO2 = 8000+round((self.listofrndms[5]*500)) #range 8000-8500)
        
        elif self.stdID == "testing":
            ###PARAMATERS for QUESTION 1
            self.newYxs = 2
            self.newH = 1.8
            self.newO = 0.5
            self.newN = 0.2
            
            ###PARAMATERS for QUESTION 2
            self.newRO2 = 3300
            self.newRCO2 = 8300
        return
    
    
###############
###############
###############
    def correct_homework_three_question_one(self, RQ):
        #Parameters:
        #RQ: Student solution. Will be comparied with opimal solution and decides if task was done correctly or not
        
        #newYxs, newH, newO, newN: changed parameters based on the seed.
            #are needed to calculate the opimal solution
        #hq1tgl: boolean. Decides if the student used the help-option or not

        #output formula:
        # X-YZ
        # X: was the help used? [yes] (randrange(5,9) or [no] randrange(0,4)
        #    -> randomint
        # YZ: Summ(studentID[0]+studentID[1]) [<10] randrange(0,4)) OR [>= 10] randrange(5,9)):
        #    -> 2x randomint
        if self.stdID == "testing":
            with self.question_one_display.children[0]:
                print('\nTesting solution calculation...\n')
                print("#####\n\nThe Std-ID is not a valid Student-ID\n and CAN'T generate a valid code!\n\n#####")

                Yxc = self.newYxs - 1
                print("\nYxc:",Yxc)
                Yxn = self.newN
                Yxw = (self.newYxs*2 + Yxn*3 - self.newH)/2
                print("\nYxw:",Yxw)
                Yxo = (-self.newYxs + Yxw + 2*Yxc + self.newO)/2
                print("\nYxo:",Yxo)

                realRQ = Yxc/Yxo
                print("\nrealRQ:",realRQ)        
        
        
        if type(self.stdID) == int:
            stdIDstr = str(self.stdID)


            #Calculation of the solution:
            with self.question_one_display.children[-1]:

                Yxc = self.newYxs - 1
                #print(Yxc)
                Yxn = self.newN
                Yxw = (self.newYxs*2 + Yxn*3 - self.newH)/2
                #print(Yxw)
                Yxo = (-self.newYxs + Yxw + 2*Yxc + self.newO)/2
                #print(Yxo)

                realRQ = Yxc/Yxo
                #print(realRQ)

                #Check if student solution is correct
                #Assumes aswer was rounded to 2 decimals
                if round(realRQ,2) == RQ:
                    self.question_one_display.children[-3].children[0].disabled = True
                    self.question_one_display.children[-3].children[1].disabled = True
                    print("Solution corect! RQ is {}".format(round(realRQ,2)))

                    solcode = ""
                     #Help-option was not used
                    if self.hq1tgl == 0:
                        solcode+=(str(rd.randrange(0,4)))
                    #Help-option was used
                    elif self.hq1tgl == 1:
                        solcode+=(str(rd.randrange(5,9)))

                    if int(stdIDstr[-1])+int(stdIDstr[-2]) < 10:
                        solcode+=(str(rd.randrange(0,4)))
                        solcode+=(str(rd.randrange(0,4)))
                    elif int(stdIDstr[-1])+int(stdIDstr[-2]) >= 10:
                        solcode+=(str(rd.randrange(5,9)))
                        solcode+=(str(rd.randrange(5,9)))    
                    # PRINT THE CODE
                    print("Here's your code!\n\n"+solcode+"\n\nPlease upload it in the designated Moodle-Task.") 

                #If the solution was not correct, the students can try again
                else:
                    print("Solution WRONG!\nDon't give up and try again!")
                    self.question_one_display.children[-1].clear_output(wait=True)



###############
###############
###############
    def correct_homework_three_question_two(self,Yield_Ethanol, Yield_Substrate):
        #Parameters:
        #Yield_Ethanol, Yield_Substrate: Student solution. Will be comparied with opimal solution and decides if task was done correctly or not
        #ewYxs, newH, newO, newN, newRO2, newRCO2: changed parameters based on the seed.
            #are needed to calculate the opimal solution
        #hq2tgl: boolean. Decides if the student used the help-option or not

        #output formula:
        # X-YZ
        # X: was the help used? yes or no
        #    -> randomint
        # YZ: Summ(studentID[0]+studentID[1]) <10 OR >= 10:
        #    -> 2x randomint

        if self.stdID == "testing":
            with self.question_two_display.children[0]:
                print('\nTesting solution calculation...\n')
                print("#####\n\nThe Std-ID is not a valid Student-ID\n and CAN'T generate a valid code!\n\n#####")


                #1. Converting biomass concentration from g/l to c-mol
                Xn = (round(46.5/(1*12+self.newH*1+self.newO*16+self.newN*14),2)*10000) 
                print("\nXn:",Xn)
                #Xn = ((50 g/L - 7 % ash)/(masses of C_1+H_newH+O_newO+N_newN) g/mol) * 10000 L = x cmol
                # 12, 1, 16, and 14 are the assumed molecular masses of the specific elements

                #2. Normalizing gas exchange rates
                rO2 = round(self.newRO2/Xn,2)
                rCO2 = round(self.newRCO2/Xn,2)
                print("\nrO2:",rO2)
                print("\nrCO2:",rCO2)

                #3. Calculating Yields, YXO, YXC
                mu = 0.35 #Growthrate = 0.35 1/h
                Yxo = round(rO2/mu,2)
                Yxc = round(rCO2/mu,2)
                Yxn = self.newN
                print("\nYxo:",Yxo)
                print("\nYxc:",Yxc)
                Ymatrixvar = np.array([[2,1],[1,-0.5]])
                Ymatrixsolvd = np.array([-1*(self.newH-2-2*Yxc-3*Yxn), -1*(self.newO+1*Yxc-1-2*Yxo)])
                Ymatrix = np.linalg.solve(Ymatrixvar,Ymatrixsolvd)
                Yxw,Yxe = Ymatrix
                Yxs = Yxe+1+Yxc
                print("\nYxe:",Yxe)
                print("\nYxs:",Yxs)

        if type(self.stdID) == int:
            stdIDstr = str(self.stdID)
            
            #Calculation of the solution:
            with self.question_two_display.children[-1]:

                #1. Converting biomass concentration from g/l to c-mol
                Xn = (round(46.5/(1*12+self.newH*1+self.newO*16+self.newN*14),2)*10000) 
                #Xn = ((50 g/L - 7 % ash)/(masses of C_1+H_newH+O_newO+N_newN) g/mol) * 10000 L = x cmol
                # 12, 1, 16, and 14 are the assumed molecular masses of the specific elements

                #2. Normalizing gas exchange rates
                rO2 = round(self.newRO2/Xn,2)
                rCO2 = round(self.newRCO2/Xn,2)

                #3. Calculating Yields, YXO, YXC
                mu = 0.35 #Growthrate = 0.35 1/h
                Yxo = round(rO2/mu,2)
                Yxc = round(rCO2/mu,2)
                Yxn = self.newN

                Ymatrixvar = np.array([[2,1],[1,-0.5]])
                Ymatrixsolvd = np.array([-1*(self.newH-2-2*Yxc-3*Yxn), -1*(self.newO+1*Yxc-1-2*Yxo)])
                Ymatrix = np.linalg.solve(Ymatrixvar,Ymatrixsolvd)
                Yxw,Yxe = Ymatrix
                Yxs = Yxe+1+Yxc

                #Check if student solution is correct
                #Assumes aswer was rounded to 2 decimals
                if round(Yxs,2) == Yield_Substrate and round(Yxe,2) == Yield_Ethanol:
                    self.question_two_display.children[-3].children[0].disabled = True
                    self.question_two_display.children[-3].children[1].disabled = True
                    print("Solution corect! Substrate yield is {}\nand Ethanol yield is {}".format(round(Yxs,2),round(Yxe,2)))

                    solcode = ""
                    #Help-option was not used
                    if self.hq2tgl == 0:
                        solcode+=(str(rd.randrange(0,4)))
                    #Help-option was used
                    elif self.hq2tgl == 1:
                        solcode+=(str(rd.randrange(5,9)))

                    if int(stdIDstr[-1])+int(stdIDstr[-2]) < 10:
                        solcode+=(str(rd.randrange(0,4)))
                        solcode+=(str(rd.randrange(0,4)))
                    elif int(stdIDstr[-1])+int(stdIDstr[-2]) >= 10:
                        solcode+=(str(rd.randrange(5,9)))
                        solcode+=(str(rd.randrange(5,9)))    
                    # PRINT THE CODE
                    print("Here's your code!\n\n"+solcode+"\n\nPlease upload it in the designated Moodle-Task.") 

                #If the solution was not correct, the students can try again
                elif round(Yxs,2) == Yield_Substrate and round(Yxe,2) != Yield_Ethanol:
                    print("Ethanol yield WRONG!\nPlease try again!")
                    self.question_two_display.children[-1].clear_output(wait=True)
                elif round(Yxs,2) != Yield_Substrate and round(Yxe,2) == Yield_Ethanol:
                    print("Substrate yield WRONG!\nPlease try again!")
                    self.question_two_display.children[-1].clear_output(wait=True)
                else:
                    #print("Real Yxe={}\nStudent_Ethanol_Yield={}".format(Yxe,Yield_Ethanol))
                    #print("Real Yxs={}\nStudent_Substrate_Yield={}".format(Yxs,Yield_Substrate))
                    print("Solution WRONG!\nDon't give up and try again!")
                    self.question_two_display.children[-1].clear_output(wait=True)

    ######
    # functions that create widgets
    #
    #
    #
    ###
    
    def build_question_one(self):
        listofwidgets = []
        
        # build question label/output (top 0)
        # build answer input (top 1)
        # build 'Check Answer button' (top 2.hbox-0)
        # build 'Get help button' (top 2. hbox-1)
        # build output (for commentsof help button) (top 3)
        # build output (for answer button feedback) (top 4)
        
        listofwidgets.append(widgets.Output(
        ))
        with listofwidgets[0]:
            #print("##########\n\nParameters for Question 1:\n\n"\
            #"Your glucose yield Y_(XS) is {} C-mol glucose/C-mol biomass\n\n"\
            #"Your biomass composition is CH_({})O_({})N_({})\n\n##########".format(self.newYxs,self.newH,self.newO,self.newN))
            print("")
            
        listofwidgets.append(widgets.FloatText(
            value=0.00,
            description='RQ:',
            disabled=False,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='200px', height='40px')
        ))


        
        i = []
        i.append(widgets.Button(
            value=False,
            description='Check RQ',
            disabled=False,
            button_style='', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Click to check solution',
            #     icon='check' # (FontAwesome names without the `fa-` prefix)
        ))
                             
        i.append(widgets.Button(
            value=False,
            description='Calculation Help',
            disabled=False,
            button_style='', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Click to get a tipp',
        #     icon='check' # (FontAwesome names without the `fa-` prefix)
        ))
                 
        listofwidgets.append(widgets.HBox(i))
        
        listofwidgets.append(widgets.Output())
        listofwidgets.append(widgets.Output())
        
        disp = widgets.VBox(listofwidgets)
        return disp
###############
###############
###############
    def build_question_two(self):
            listofwidgets = []

            # build question label/output (top 0)
            # build answer input for Ethanol Yield (top 1)
            # build answer input for Substrate Yield (top 2)
            # build 'Check Answer button' (top 3.hbox-0)
            # build 'Get help button' (top 3. hbox-1)
            # build output (for commentsof help button) (top 4)
            # build output (for answer button feedback) (top 5)

            listofwidgets.append(widgets.Output(
            ))
            with listofwidgets[0]:
                #print("##########\n\nParameters for Question 2:\n\n"\
                #"Your O_2 Exchange-Rate is {} mol/h\n\n"\
                #"Your CO_2 Exchange-Rate is {} mol/h\n\n##########".format(self.newRO2,self.newRCO2))
                print("")
                
            listofwidgets.append(widgets.FloatText(
                value=0.00,
                description='Ethanol Yield:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='200px', height='40px')
            ))
            
            listofwidgets.append(widgets.FloatText(
                value=0.00,
                description='Substrate Yield:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='200px', height='40px')
            ))


            i = []
            i.append(widgets.Button(
                value=False,
                description='Check Yields',
                disabled=False,
                button_style='', # 'success', 'info', 'warning', 'danger' or ''
                tooltip='Click to check solution',
                #     icon='check' # (FontAwesome names without the `fa-` prefix)
            ))

            i.append(widgets.Button(
                value=False,
                description='Calculation Help',
                disabled=False,
                button_style='', # 'success', 'info', 'warning', 'danger' or ''
                tooltip='Click to get a tipp',
            #     icon='check' # (FontAwesome names without the `fa-` prefix)
            ))

            listofwidgets.append(widgets.HBox(i))

            listofwidgets.append(widgets.Output())
            listofwidgets.append(widgets.Output())

            disp = widgets.VBox(listofwidgets)
            return disp
    
    ######
    # ON-CLICK functions
    # 2 per question
    #
    #
    ###
    def on_question_one_help_button_clicked(self, click):
        self.hq1tgl = 1
        self.question_one_display.children[-3].children[1].disabled = True
        with self.question_one_display.children[-2]:
            txt = "##########\n" \
            "* The RQ is rate of CO_2 production divivided by the rate of O_2 production\n" \
            "  RQ = r(CO_2)/r(O_2)\n\n" \
            "* Yields correlate directly to rates\n" \
            "  RQ = r(CO_2)/r(O_2) = Y(CO_2)/Y(O_2)\n\n" \
            "* By calculating all yields for the known reaction, you will be able to identify the required yields\n" \
            "##########"
            print(txt)
    def on_question_one_answer_button_clicked(self, click):
        student_answer = self.question_one_display.children[1].value
        self.correct_homework_three_question_one(student_answer)
        
    def on_question_two_help_button_clicked(self, click):
        self.hq2tgl = 1
        self.question_two_display.children[-3].children[1].disabled = True
        with self.question_two_display.children[-2]:
            txt = "##########\n" \
            "Don't forget, that for Question 2, Ethanol is an additional product!\nYou also need to subtract the ash from the assumed Biomass.\n\n"\
            "* First, convert the biomass concentration from g/L to c-mol\n" \
            "  and normalize the gas exchange rates\n\n" \
            "* You can use the growth rates and exchange rates to determine\n the yields of O_2 and CO_2\n\n" \
            "* How to solve the matrix can be seen in 'Example 1' (p. 19/20 of Lecture QM3 ‘Quantitative Microbiologie week 4 – Folien’)\n"
            "##########\n"
            print(txt)
    def on_question_two_answer_button_clicked(self, click):
        student_answer_ethanol_yield = self.question_two_display.children[1].value
        student_answer_substrate_yield = self.question_two_display.children[2].value
        self.correct_homework_three_question_two(student_answer_ethanol_yield, student_answer_substrate_yield)

##########
##########
##########
#class for homework 4
##########
##########
##########

class homework_four():
#class homework_four_test():
    def __init__(self, stdID):
        # Check if Student-ID is an integer and if the function is to be run in "test-mode"
        
        #'TEst-Mode' is activated by the Std-ID input: 123456789
        # For this to be viable, 123456789 CAN NEVER be a viable student-id
        if type(stdID) == int and stdID == 123456789:
            self.stdID = "testing"
        elif type(stdID) == int and stdID != 123456789:
            self.stdID = stdID
        else:
            print("Student ID required to be an integer.\nPlease try again!\n\n")
            pass
        
        # How many rngs are to be created for new/changed parameters
        self.numberofrandomnumbers = 8
        
        # Create RNs on the basis of the studentID
        self.initiate_homework()
        
        # Create variables using the generated RNs
        self.create_question_variables()
        
        # List of self.parameters:
        #self.stID
        #self.listofrndms
        
        #self.newCoP (CombustionofProtein)
        #self.newCoC (CombustionofCarbohydrates)
        #self.newCoF (CombustionofFat)
        
        #self.newOCR (OxygenConsumptionrate)
        #self.newCPR (Carbon dioxide production rate)
        #self.newNSR (Nitrogen secretion rate)
        
        self.hq1tgl = 0 #help_question_1_toggle
        self.hq2tgl = 0 #help_question_2_toggle
        self.hq3tgl = 0 #help_question_2_toggle
        
        # build widgets for the questions
        self.question_one_display = self.build_question_one()
        self.question_two_display = self.build_question_two()
        self.question_three_display = self.build_question_three()
        
        
    def initiate_homework(self):
    # Parameters:
    # stdID is an integer used to seed the rng
        # IF stdID is given the string "test", it's supposed to return the string "test"
        # This string is then used to thest functions using standard values
    # counter is an integer used to define how many rngs are created.
        #print("Matriculation-Number accepted.\n\nGenerating random numbers..")
        #print("Matriculation-Number accepted.\n\nGenerating personalized homework parameters..")
        if self.stdID == "testing":
            print("#####\n\nTesting functionality\n\n#####")
            print("#####\n\The Std-ID is not a valid Student-ID\n and CAN'T generate a valid code!\n\n#####")
            # NOTE: no self.listofrndms is created in test mode
        else:
        #RNG for this homework
        #IMPORTANT: rd.seed() is set GLOBALLY, as far as we observed
        #this means: the solution-codes are ALWAYS IDENTICAL FOR A UNIQUE STid
        #if rd.seed() is not refreshed after generation of listofrndms.
        #In this version, the generated solution-codes are supposed to be random, therefore. rd.seed() gets refreshed
            rd.seed(self.stdID)
            self.listofrndms = [rd.random() for i in range(self.numberofrandomnumbers)] #create new random numbers that can be used to modify parameters
            rd.seed()
        return
    
    def create_question_variables(self):
        #function returns all generated homework parameters and a boolean 
        #thats used to toggls help_question_1_toggle to  0 (helpoption was not used)
        if self.stdID != "testing":
            #print("\nGenerating personalized homework parameters...\n")
            ###PARAMATERS for QUESTION 1
            self.newCoP = 4.0+round((self.listofrndms[0]*0.2),1) #range 4.0-4.2
            self.newCoC = 4.0+round((self.listofrndms[1]*0.3),1) #range 4.0-4.3
            self.newCoF = 9.0+round((self.listofrndms[2]*0.6),1) #range 9.0-9.6

            ###PARAMATERS for QUESTION 3
            self.newOCR = 0.5+round((self.listofrndms[4]*0.2)) #range 0.5-0.7)
            self.newCPR = 0.4+round((self.listofrndms[5]*0.2)) #range 0.4-0.6)
            self.newNSR = 0.05+round((self.listofrndms[6]*0.2)) #range 0.05-0.25)
            
        elif self.stdID == "testing":
            ###PARAMATERS for QUESTION 1
            self.newCoP = 4.1 #(CombustionofProtein [kcal/g])
            self.newCoC = 4.2 #(CombustionofCarbohydrates [kcal/g])
            self.newCoF = 9.3 #(CombustionofFat [kcal/g])
            
           ###PARAMATERS for QUESTION 3
            self.newOCR = 0.6 #(OxygenConsumptionrate [mol/h])
            self.newCPR = 0.52 #(Carbon dioxide production rate [mol/h])
            self.newNSR = 0.1 #(Nitrogen secretion rate [N/h])

        return
    
    
###############
###############
###############
    def correct_homework_four_question_one(self, YxqP, YxqC, YxqF):
        #Parameters:
        #YxqP, YxqC, YxqF: Student solutions. Will be comparied with opimal solution and decides if task was done correctly or not
        
        #newCoP, newCoC, newCoF: changed parameters based on the seed.
            #are needed to calculate the opimal solution
        #hq1tgl: boolean. Decides if the student used the help-option or not

        #output formula:
        # X-YZ
        # X: was the help used? yes or no
        #    -> randomint
        # YZ: Summ(studentID[0]+studentID[1]) <10 OR >= 10:
        #    -> 2x randomint
        if self.stdID == "testing":
            with self.question_one_display.children[0]:
                print('\nTesting solution calculation...\n')
                print("#####\n\nThe Std-ID is not a valid Student-ID\n and CAN'T generate a valid code!\n\n#####")

                mp = 12+1.57+16*0.32+14*0.26 #CH_1.57O_0.32N_0.26
                mc = 12+2+16 #CH_2O
                mf = 12+1.92+16*0.12 #CH_1.92O_0.12
                
                print(self.newCoP * mp)
                print(self.newCoC * mc)
                print(self.newCoF * mf)
        
        
        if type(self.stdID) == int:
            stdIDstr = str(self.stdID)


            #Calculation of the solution:
            with self.question_one_display.children[-1]:

                mp = 12+1.57+16*0.32+14*0.26 #CH_1.57O_0.32N_0.26
                mc = 12+2+16 #CH_2O
                mf = 12+1.92+16*0.12 #CH_1.92O_0.12
            
                #Check if student solutions are correct
                counter = 0
                #Assumes aswer was rounded to 2 decimals
                if round(mp*self.newCoP,2) != YxqP:
                    counter += 1
                    print("Wrong Combustion of Protein\n")
                if round(mc*self.newCoC,2) != YxqC:
                    counter += 1
                    print("Wrong Combustion of Carbohydrates\n")
                if round(mf*self.newCoF,2) != YxqF:
                    counter += 1
                    print("Wrong Combustion of Fat\n")
                if round(mp*self.newCoP,2) == YxqP and round(mc*self.newCoC,2) == YxqC and round(mf*self.newCoF,2) == YxqF:
                    self.question_one_display.children[-3].children[0].disabled = True
                    self.question_one_display.children[-3].children[1].disabled = True
                    print("Solution corect!")

                    solcode = ""
                    #Help-option was not used
                    if self.hq1tgl == 0:
                        solcode+=(str(rd.randrange(0,4)))
                    #Help-option was used
                    elif self.hq1tgl == 1:
                        solcode+=(str(rd.randrange(5,9)))

                    if int(stdIDstr[-1])+int(stdIDstr[-2]) < 10:
                        solcode+=(str(rd.randrange(0,4)))
                        solcode+=(str(rd.randrange(0,4)))
                    elif int(stdIDstr[-1])+int(stdIDstr[-2]) >= 10:
                        solcode+=(str(rd.randrange(5,9)))
                        solcode+=(str(rd.randrange(5,9)))    
                    # PRINT THE CODE
                    print("Here's your code!\n\n"+solcode+"\n\nPlease upload it in the designated Moodle-Task.") 

                if counter != 0:
                    print("\n##########\n Don't give up and try again!\n\n")
                    self.question_one_display.children[-1].clear_output(wait=True)


###############
###############
###############
    def correct_homework_four_question_two(self,student_answer_Ycs, student_answer_Ycf, student_answer_Ycp):
        #Parameters:
        #Yield_sugar (student_answer_Ycs), Yield_fat: student_answer_Ycf, Yield_protein student_answer_Ycp
        
        #newCoP, newCoC, newCoF: changed parameters based on the seed.
        #newOCR, newCPR, newNSR: changed parameters based on the seed.
            #are needed to calculate the opimal solution
        #hq2tgl: boolean. Decides if the student used the help-option or not

        #output formula:
        # X-YZ
        # X: was the help used? yes or no
        #    -> randomint
        # YZ: Summ(studentID[0]+studentID[1]) <10 OR >= 10:
        #    -> 2x randomint
        
        
        # Hopefully temporary: list of solutions for all three Yields
        # In this HW_4 section, no parameters were changed yet.
        # Thus, the literature solutions arethe correct solutions
        #format: 
        #Y_cs = {A} + {B}*Y_co + {C}*Y_cn"
        #sol_Ycs= [A, B, C]
        sol_Ycs = [-2.38,3.38,-3.58]
        sol_Ycf = [2.38,-2.38,-0.27]
        sol_Ycp = [0.0,0.0,3.85]
        
        
        if self.stdID == "testing":
            with self.question_two_display.children[0]:
                print('\nTesting solution calculation...\n')
                print("#####\n\nThe Std-ID is not a valid Student-ID\n and CAN'T generate a valid code!\n\n#####")
                
                print("Yield sugar/carbohydrates: {} + {}*Y_co + {}*Y_c\n\n".format(sol_Ycs[0],sol_Ycs[1],sol_Ycs[2]))
                print("Yield fat: {} + {}*Y_co + {}*Y_c\n\n".format(sol_Ycf[0],sol_Ycf[1],sol_Ycf[2]))
                print("Yield protein: {} + {}*Y_co + {}*Y_c\n\n".format(sol_Ycp[0],sol_Ycp[1],sol_Ycp[2]))

        if type(self.stdID) == int:
            stdIDstr = str(self.stdID)
            
            #Calculation of the solution:
            with self.question_two_display.children[-1]:
                counter = 0
                
                print("Y_cs: {1} + {2} * Y_co + {3} * Y_cn\n-----------------------------\n")
                for i in range(len(student_answer_Ycs)):
                    if round(student_answer_Ycs[i],1) != round(sol_Ycs[i],1):
                        counter += 1
                        print('Variable {} is WRONG\n'.format(i+1))
                print("Y_cf: {1} + {2} * Y_co + {3} * Y_cn\n-----------------------------\n")
                for i in range(len(student_answer_Ycf)):
                    if round(student_answer_Ycf[i],1) != round(sol_Ycf[i],1):
                        counter += 1
                        print('Variable {} is WRONG\n'.format(i+1))
                print("Y_cp: {1} + {2} * Y_co + {3} * Y_cn\n-----------------------------\n")
                for i in range(len(student_answer_Ycp)):
                    if round(student_answer_Ycp[i],1) != round(sol_Ycp[i],1):
                        counter += 1
                        print('Variable {} is WRONG\n'.format(i+1))
                if counter != 0:
                    print("\n##########\n Don't give up and try again!\n\n")
                    self.question_two_display.children[-1].clear_output(wait=True)
                elif counter == 0:
                    self.question_two_display.children[-3].children[0].disabled = True
                    self.question_two_display.children[-3].children[1].disabled = True
                    print("Solution correct!")

                    solcode = ""
                    #Help-option was not used
                    if self.hq2tgl == 0:
                        solcode+=(str(rd.randrange(0,4)))
                    #Help-option was used
                    elif self.hq2tgl == 1:
                        solcode+=(str(rd.randrange(5,9)))

                    if int(stdIDstr[-1])+int(stdIDstr[-2]) < 10:
                        solcode+=(str(rd.randrange(0,4)))
                        solcode+=(str(rd.randrange(0,4)))
                    elif int(stdIDstr[-1])+int(stdIDstr[-2]) >= 10:
                        solcode+=(str(rd.randrange(5,9)))
                        solcode+=(str(rd.randrange(5,9)))    
                    # PRINT THE CODE
                    print("Here's your code!\n\n"+solcode+"\n\nPlease upload it in the designated Moodle-Task.")

###############
###############
###############
    def correct_homework_four_question_three(self,student_answer_Yco, student_answer_Ycn, student_answer_Rxq):
        #Parameters:
        #Yield_co (student_answer_Yco) normalized to CO2, Yield_cn: student_answer_Ycn normalized to CO2, metablic rate student_answer_Rxq (kcal/h)
        
        #newCoP, newCoC, newCoF: changed parameters based on the seed.
        #newOCR, newCPR, newNSR: changed parameters based on the seed.
            #are needed to calculate the opimal solution
        #hq3tgl: boolean. Decides if the student used the help-option or not
        
        #ALSO:
        # Instead of a literature solution matrix, the student solution of Question 2 is to be used as a basis
        # So that they can get points for Question 3, even if they got Question 2 wrong

        #output formula:
        # X-YZ
        # X: was the help used? yes or no
        #    -> randomint
        # YZ: Summ(studentID[0]+studentID[1]) <10 OR >= 10:
        #    -> 2x randomint
        
        
        # Hopefully temporary: list of solutions for all three Yields
        # In this HW_4 section, te parameters of the chemical composition parameters were not changed yet.
        # Thus, the literature solutions are the correct solutions of Homework_4_2:
        #format: 
        #Y_cs = {A} + {B}*Y_co + {C}*Y_cn"
        #sol_Ycs= [A, B, C]
        #sol_Ycs = [-2.38,3.38,-3.58]
        #sol_Ycf = [2.38,-2.38,-0.27]
        #sol_Ycp = [0.0,0.0,3.85]     
        
        mp = 12+1.57+16*0.32+14*0.26 #CH_1.57O_0.32N_0.26
        mc = 12+2+16 #CH_2O
        mf = 12+1.92+16*0.12 #CH_1.92O_0.12
        
        if self.stdID == "testing":
            with self.question_three_display.children[0]:
                print('\nTesting solution calculation...\n')
                print("#####\n\nThe Std-ID is not a valid Student-ID\n and CAN'T generate a valid code!\n\n#####")
                
                Ycc = round(self.newCPR/self.newCPR,2) #(mol/h / mol/h)
                Yco = round(self.newOCR/self.newCPR,2) #(mol/h / mol/h)
                rn = self.newNSR/14 #(Secretion rate NH3 in gN/h / 14 g/mol)
                Ycn = round(rn/self.newCPR,2) #(mol/h / mol/h)
                
                print('Yco: \n',Yco)
                print('Ycn: \n',Ycn)
                
                Ycs = round(sol_Ycs[0]*Ycc+sol_Ycs[1]*Yco+sol_Ycs[2]*Ycn,2) #(mol/cmol)
                Ycf = round(sol_Ycf[0]*Ycc+sol_Ycf[1]*Yco+sol_Ycf[2]*Ycn,2) #(mol/cmol)
                Ycp = round(sol_Ycp[0]*Ycc+sol_Ycp[1]*Yco+sol_Ycp[2]*Ycn,2) #(mol/cmol)
                
                print('Ycs: \n',Ycs)
                print('Ycf: \n',Ycf)
                print('Ycp: \n',Ycp)
                print('mc: \n', mc)
                print('mf: \n', mf)
                print('mp: \n', mp)
                
                Yxq = Ycs*round(mc*self.newCoC,2)+Ycf*round(mf*self.newCoF,2)+Ycp*round(mp*self.newCoP,2) #(mol/cmol * kcal/cmol) = (kcal/cmol) || units not clear, might need second look
                Rxq = round(Yxq * self.newCPR,2) # kcal/cmol * mol/h = kcal/h || units not clear, might need second look
                
                print('Yxq: \n',Yxq)
                print('Rxq: \n',Rxq)


        if type(self.stdID) == int:
            stdIDstr = str(self.stdID)
            
            #Calculation of the solution:
            with self.question_three_display.children[-1]:
                
                # Instead of a literature solution, the student solution is to be used as a basis
                # So that they can get points for Question 3, even if they got Question 2 wrong
                sol_Ycs = []
                sol_Ycf = []
                sol_Ycp = []
                for i in self.question_two_display.children[2].children:
                    sol_Ycs.append(i.value)
                for i in self.question_two_display.children[4].children:
                    sol_Ycf.append(i.value)
                for i in self.question_two_display.children[6].children:
                    sol_Ycp.append(i.value)
                #RUDIMENTARY check if there are problems with the student solution of Question 2
                checkcounter = 0
                if sum(sol_Ycs) == 0:
                    checkcounter += 1
                if sum(sol_Ycf) == 0:
                    checkcounter += 1
                if sum(sol_Ycp) == 0:
                    checkcounter += 1
                    
                if checkcounter > 0:
                    print("\n##########\nNo Solution of Question 2 detected!\nPlease try to solve Question 2 before answering Question 3!\n\n")
                    self.question_three_display.children[-1].clear_output(wait=True)
                    return
                    
                #if Question 2 answers are accepted, continue here    
                Ycc = round(self.newCPR/self.newCPR,2) #(mol/h / mol/h)
                Yco = round(self.newOCR/self.newCPR,2) #(mol/h / mol/h)
                rn = self.newNSR/14 #(Secretion rate NH3 in gN/h / 14 g/mol)
                Ycn = round(rn/self.newCPR,2) #(mol/h / mol/h)
                
                #Check first Checkpoint
                counter = 0
                if Yco != student_answer_Yco:
                    counter += 1
                    print("Normalized Yield_co WRONG!\n")
                if Ycn != student_answer_Ycn:
                    counter += 1
                    print("Normalized Yield_cn WRONG!\n")

                       
                Ycs = round(sol_Ycs[0]*Ycc+sol_Ycs[1]*Yco+sol_Ycs[2]*Ycn,2) #(mol/cmol)
                Ycf = round(sol_Ycf[0]*Ycc+sol_Ycf[1]*Yco+sol_Ycf[2]*Ycn,2) #(mol/cmol)
                Ycp = round(sol_Ycp[0]*Ycc+sol_Ycp[1]*Yco+sol_Ycp[2]*Ycn,2) #(mol/cmol)
                

                Yxq = Ycs*round(mc*self.newCoC,2)+Ycf*round(mf*self.newCoF,2)+Ycp*round(mp*self.newCoP,2) #(mol/cmol * kcal/cmol) = (kcal/cmol) || units not clear, might need second look
                Rxq = round(Yxq * self.newCPR,2) # kcal/cmol * mol/h = kcal/h || units not clear, might need second look
                
                #Check second Checkpoint
                if Rxq != student_answer_Rxq:
                    counter += 1
                    print("Metabolic rate WRONG!\n")
                
                
                if counter != 0:
                    print("\n##########\n Don't give up and try again!\n\n")
                    self.question_three_display.children[-1].clear_output(wait=True)
                elif counter == 0:
                    self.question_three_display.children[-3].children[0].disabled = True
                    self.question_three_display.children[-3].children[1].disabled = True
                    print("Solution corect!")
                    solcode = ""
                    #Help-option was not used
                    if self.hq3tgl == 0:
                        solcode+=(str(rd.randrange(0,4)))
                    #Help-option was used
                    elif self.hq3tgl == 1:
                        solcode+=(str(rd.randrange(5,9)))

                    if int(stdIDstr[-1])+int(stdIDstr[-2]) < 10:
                        solcode+=(str(rd.randrange(0,4)))
                        solcode+=(str(rd.randrange(0,4)))
                    elif int(stdIDstr[-1])+int(stdIDstr[-2]) >= 10:
                        solcode+=(str(rd.randrange(5,9)))
                        solcode+=(str(rd.randrange(5,9)))    
                    # PRINT THE CODE
                    print("Here's your code!\n\n"+solcode+"\n\nPlease upload it in the designated Moodle-Task.")
                

    ######
    # functions that create widgets
    #
    #
    #
    ###
    
    def build_question_one(self):
        listofwidgets = []
        
        # build question label/output (top 0)
        # build answer input YxqP(top 1)
        # build answer input YxqC(top 2)
        # build answer input YxqF(top 3)
        # build 'Check Answer button' (top 4.hbox-0)
        # build 'Get help button' (top 4. hbox-1)
        # build output (for commentsof help button) (top 5)
        # build output (for answer button feedback) (top 6)
        
        listofwidgets.append(widgets.Output(
        ))
        with listofwidgets[0]:
            #print("##########\n\nParameters for Question 1:\n\n"\
            #"Your glucose yield Y_(XS) is {} C-mol glucose/C-mol biomass\n\n"\
            #"Your biomass composition is CH_({})O_({})N_({})\n\n##########".format(self.newYxs,self.newH,self.newO,self.newN))
            print("")
        
        listofwidgets.append(widgets.FloatText(
            value=0.00,
            description='YxqP [kcal/mol]',
            disabled=False,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='200px', height='40px')
        ))

        listofwidgets.append(widgets.FloatText(
            value=0.00,
            description='YxqC [kcal/mol]',
            disabled=False,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='200px', height='40px')
        ))
        
        listofwidgets.append(widgets.FloatText(
            value=0.00,
            description='YxqF [kcal/mol]',
            disabled=False,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='200px', height='40px')
        ))
        
        i = []
        i.append(widgets.Button(
            value=False,
            description='Check Solution',
            disabled=False,
            button_style='', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Click to check solution',
            #     icon='check' # (FontAwesome names without the `fa-` prefix)
        ))
                             
        i.append(widgets.Button(
            value=False,
            description='Calculation Help',
            disabled=False,
            button_style='', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Click to get a tipp',
        #     icon='check' # (FontAwesome names without the `fa-` prefix)
        ))
        
        listofwidgets.append(widgets.HBox(i))
        
        listofwidgets.append(widgets.Output())
        listofwidgets.append(widgets.Output())
        
        disp = widgets.VBox(listofwidgets)
        return disp
###############
###############
###############
    def build_question_two(self):
            listofwidgets = []

            # build question label/output (top 0)
            # build answer format label for Ycs (top 1)
                #build answer input for Ycs answer {A} (top 2.hbox-0)
                #build answer input for Ycs answer {B} (top 2.hbox-1)
                #build answer input for Ycs answer {C} (top 2.hbox-2)
            # build answer input for Ycf (top 3)
                #build answer input for Ycf answer {A} (top 4.hbox-0)
                #build answer input for Ycf answer {B} (top 4.hbox-1)
                #build answer input for Ycf answer {C} (top 4.hbox-2)
            # build answer input for Ycn (top 5)
                #build answer input for Ycn answer {A} (top 6.hbox-0)
                #build answer input for Ycn answer {B} (top 6.hbox-1)
                #build answer input for Ycn answer {C} (top 6.hbox-2)
            # build 'Check Answer button' (top 7.hbox-0)
            # build 'Get help button' (top 7. hbox-1)
            # build output (for commentsof help button) (top 8)
            # build output (for answer button feedback) (top 9)

            listofwidgets.append(widgets.Output(
            ))
            with listofwidgets[0]:
                #print("##########\n\nParameters for Question 2:\n\n"\
                #"Your O_2 Exchange-Rate is {} mol/h\n\n"\
                #"Your CO_2 Exchange-Rate is {} mol/h\n\n##########".format(self.newRO2,self.newRCO2))
                print("")
            ###
            # Widgtes responsible for student input of Y_cs
            ###
            listofwidgets.append(widgets.Label(
                value="Yield: Y_cs [mol/cmol]= {A} + {B}*Y_co + {C}*Y_cn",
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='500px', height='40px')
            ))
            lstOne = []
            lstOne.append(widgets.FloatText(
                value=0.0,
                description='A:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='100px', height='40px')
            ))
            lstOne.append(widgets.FloatText(
                value=0.0,
                description='B:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='100px', height='40px')
            ))
            lstOne.append(widgets.FloatText(
                value=0.0,
                description='C:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='100px', height='40px')
            ))
            listofwidgets.append(widgets.HBox(lstOne))

            ###
            # Widgtes responsible for student input of Y_cf
            ###
            listofwidgets.append(widgets.Label(
                value="Yield: Y_cf [mol/cmol]= {A} + {B}*Y_co + {C}*Y_cn",
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='500px', height='40px')
            ))
            lstTwo = []
            lstTwo.append(widgets.FloatText(
                value=0.0,
                description='A:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='100px', height='40px')
            ))
            lstTwo.append(widgets.FloatText(
                value=0.0,
                description='B:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='100px', height='40px')
            ))
            lstTwo.append(widgets.FloatText(
                value=0.0,
                description='C:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='100px', height='40px')
            ))
            listofwidgets.append(widgets.HBox(lstTwo))
            
            ###
            # Widgtes responsible for student input of Y_cp
            ###
            listofwidgets.append(widgets.Label(
                value="Yield: Y_cp [mol/cmol]= {A} + {B}*Y_co + {C}*Y_cn",
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='500px', height='40px')
            ))
            lstThree = []
            lstThree.append(widgets.FloatText(
                value=0.0,
                description='A:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='100px', height='40px')
            ))
            lstThree.append(widgets.FloatText(
                value=0.0,
                description='B:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='100px', height='40px')
            ))
            lstThree.append(widgets.FloatText(
                value=0.0,
                description='C:',
                disabled=False,
                display='flex',
                flex_flow='column',
                align_items='stretch',
                style= {'description_width': 'initial'},
                layout = widgets.Layout(width='100px', height='40px')
            ))
            listofwidgets.append(widgets.HBox(lstThree))
            
            
            ###
            # Wisgets resonsible for Solution check and help button
            #
            i = []
            i.append(widgets.Button(
                value=False,
                description='Check Yields',
                disabled=False,
                button_style='', # 'success', 'info', 'warning', 'danger' or ''
                tooltip='Click to check solution',
                #     icon='check' # (FontAwesome names without the `fa-` prefix)
            ))

            i.append(widgets.Button(
                value=False,
                description='Calculation Help',
                disabled=False,
                button_style='', # 'success', 'info', 'warning', 'danger' or ''
                tooltip='Click to get a tipp',
            #     icon='check' # (FontAwesome names without the `fa-` prefix)
            ))

            listofwidgets.append(widgets.HBox(i))

            listofwidgets.append(widgets.Output())
            listofwidgets.append(widgets.Output())

            disp = widgets.VBox(listofwidgets)
            return disp
        
###############
###############
###############        
    def build_question_three(self):
        listofwidgets = []
        
        # build question label/output (top 0)
        # build answer input Yco(top 1)
        # build answer input Ycn(top 2)
        # build answer input Rxq(top 3)
        # build 'Check Answer button' (top 4.hbox-0)
        # build 'Get help button' (top 4. hbox-1)
        # build output (for commentsof help button) (top 5)
        # build output (for answer button feedback) (top 6)
        
        listofwidgets.append(widgets.Output(
        ))
        with listofwidgets[0]:
            print("")
        
        listofwidgets.append(widgets.FloatText(
            value=0.00,
            description='Yco [mol/cmol]',
            disabled=False,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='200px', height='40px')
        ))

        listofwidgets.append(widgets.FloatText(
            value=0.00,
            description='Ycn [mol/cmol]',
            disabled=False,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='200px', height='40px')
        ))
        
        listofwidgets.append(widgets.FloatText(
            value=0.00,
            description='Rxq [kcal/h]',
            disabled=False,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='200px', height='40px')
        ))
        
        i = []
        i.append(widgets.Button(
            value=False,
            description='Check Solution',
            disabled=False,
            button_style='', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Click to check solution',
            #     icon='check' # (FontAwesome names without the `fa-` prefix)
        ))
                             
        i.append(widgets.Button(
            value=False,
            description='Calculation Help',
            disabled=False,
            button_style='', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Click to get a tipp',
        #     icon='check' # (FontAwesome names without the `fa-` prefix)
        ))
        
        listofwidgets.append(widgets.HBox(i))
        
        listofwidgets.append(widgets.Output())
        listofwidgets.append(widgets.Output())
        
        disp = widgets.VBox(listofwidgets)
        return disp
    
    ######
    # ON-CLICK functions
    # 2 per question
    #
    #
    ###
    def on_question_one_help_button_clicked(self, click):
        self.hq1tgl = 1
        self.question_one_display.children[-3].children[1].disabled = True
        with self.question_one_display.children[-2]:
            txt = "##########\n" \
            "Given are the estimated heat of combustions in (kcal/g) and\nthe chemical composition.\n\n" \
            "* Sum up the approximate weight of each subgroup in g/cmol\n\n" \
            "* Multiply the 'heat of combustion' (kcal/g) with the 'molecular weight' (g/mol)\n" \
            "##########"
            print(txt)
    def on_question_one_answer_button_clicked(self, click):
        student_YxqP = self.question_one_display.children[1].value
        student_YxqC = self.question_one_display.children[2].value
        student_YxqF = self.question_one_display.children[3].value
        self.correct_homework_four_question_one(student_YxqP, student_YxqC, student_YxqF)
        
    def on_question_two_help_button_clicked(self, click):
        self.hq2tgl = 1
        self.question_two_display.children[-3].children[1].disabled = True
        with self.question_two_display.children[-2]:
            txt ="""
##########\n
Given is the stoichiometry of the combustion of Carbohydrates (Ycs), Fats (Ycf), and Protein (Ycp)\ntogether with Oxygen (Yco) to form CO_2 (Ycc), ammonia (Ycn) and water (Ycw).\n
The equation is normalized on CO_2, so the yield of CO_2 (Ycc) = 1.\n
The task is to solve Ycs, Ycf and Ycp as functions of Yco and Ycn.\n
One way to do this, is to calculate the degree of reduction (DoR) as can be seen on p. 23-25 in the\nQM3-script ‘Quantitative Microbiologie week 4 – Folien’.\n
Setting up a stoichiometric matrix should look like this:\n\n
\t\t\t\t\t\t\t\tYcs\n
\t\t\t\t\t\t\t\tYcf\n
C  :\t1\t1\t1\t0\t0\t1\t\tYcp\n
N  :\t0\t0\tx\t0\t1\t0\t*\tYco\t=\t0\n
DoR:\tx\tx\tx\t4\t0\t0\t\tYcn\n
\t\t\t\t\t\t\t\tYcc\n\n

And rearranging the matrix using Gaussian elimination to get functions of Yco and Ycn should end up looking like this:\n\n
\t\t\t\t\t\t\t\tYcs\n
\t\t\t\t\t\t\t\tYcf\n
C  :\t1\t0\t0 \tB_cs\tC_cs\tA_cs\t\tYcp\n
N  :\t0\t1\t0\tB_cf\tC_cf\tA_cf\t*\tYco\t=\t0\n
DoR:\t0\t0\t1\tB_cp\tC_cp\tA_cp\t\tYcn\n
\t\t\t\t\t\t\t\tYcc\n\n
Meaning that the example-equation of Ycs looks like this: \n
Ycs + B_cs*Yco + C_cs*Ycn + A_cs*Ycc(with Ycc=1) = 0\n
Rearrange it for Ycs and repeat the procedure for Ycf and Ycn.\n
##########\n
            """

            print(txt)
            
    def on_question_two_answer_button_clicked(self, click):
        student_answer_Ycs = []
        student_answer_Ycf = []
        student_answer_Ycp = []
        for i in self.question_two_display.children[2].children:
            student_answer_Ycs.append(i.value)
        for i in self.question_two_display.children[4].children:
            student_answer_Ycf.append(i.value)
        for i in self.question_two_display.children[6].children:
            student_answer_Ycp.append(i.value)
        self.correct_homework_four_question_two(student_answer_Ycs, student_answer_Ycf, student_answer_Ycp)
        
    #def on_question_two_reset_button_clicked(self, click):
    #    #A button that, when clicked resets the value fo the input frlds to the original format, thus showing how the input has to look like
    #    self.question_two_display.children[1].value = "{A} + {B}*Y_co + {C}*Y_cn" #input for Ycs (top 1)
    #    self.question_two_display.children[2].value = "{A} + {B}*Y_co + {C}*Y_cn" #input for Ycf (top 1)
    #    self.question_two_display.children[3].value = "{A} + {B}*Y_co + {C}*Y_cn" #input for Ycn (top 1)
    #    with self.question_two_display.children[-1]:
    #        print("#####\nInput-Fields RESET\n#####")
    #        self.question_two_display.children[-1].clear_output(wait=True)
    
    def on_question_three_answer_button_clicked(self, click):
        student_Yco = self.question_three_display.children[1].value
        student_Ycn = self.question_three_display.children[2].value
        student_Rxq = self.question_three_display.children[3].value
        self.correct_homework_four_question_three(student_Yco, student_Ycn, student_Rxq)
    
    def on_question_three_help_button_clicked(self, click):
        self.hq3tgl = 1
        self.question_three_display.children[-3].children[1].disabled = True
        with self.question_three_display.children[-2]:
            txt ="""
##########\n
Task is determine the metabolic rate Rxq [kcal/h]\n
 * First step is to calculate the Yields of O_2 (Yco) and NH_4 (Ycn)relative to CO_2\n
(* The Yied of CO_2 (Y_cc) = 1 [mol/cmol])\n
(* Don’t forget to calculate Ycn out of the nitrogen secretion rate!)\n
 * Next, enter the yields into the three equations from [Question 2].\n
 * Calculate Yxq [kcal/cmol] (p. 36-38  in the QM3-script ‘Quantitative Microbiologie week 4 – Folien’\n
 * The formula for Rxq is Yxq [kcal/cmol] * r_co2 [cmol/h]\n
##########\n
            """

            print(txt)

##########
##########
##########
#class for homework 5
##########
##########
##########

class homework_five():
#class homework_five_test():
    def __init__(self):
        self.init_display = self.build_init_display()
        
        
        
    def start_homework(self, stdID):
        # Check if Student-ID is an integer and if the function is to be run in "test-mode"
        
        #'TEst-Mode' is activated by the Std-ID input: 123456789
        # For this to be viable, 123456789 CANNEVER be a viable student-id
        if type(stdID) == int and stdID == 123456789:
            self.stdID = "testing"
        elif type(stdID) == int and stdID != 123456789:
            self.stdID = stdID
        else:
            print("Student ID required to be an integer.\nPlease try again!\n\n")
            pass
        
        # How many rngs are to be created for new/changed parameters
        self.numberofrandomnumbers = 10
        
        # Create RNs on the basis of the studentID
        self.initiate_homework()
        
        # Create variables using the generated RNs
        self.create_question_variables()
        
        # List of self.parameters:
        #self.stID
        #self.listofrndms
        
        #self.v1
        #self.v4
        #self.v5
        #self.v6
        #self.v7
        
        self.hq1tgl = 0 #help_question_1_toggle
        self.hq2tgl = 0 #help_question_2_toggle
        
        # build widgets for the questions
        self.question_one_display = self.build_question_display()
        self.question_two_display = self.build_question_display()
        
        with self.init_display.children[2]:
            txt= """
##########\n\nParameters for Question 2:\n\n
Flux v1 = {}\n
Flux v4 = {}\n
Flux v5 = {}\n
Flux v6 = {}\n
Flux v7 = {}\n\n
##########
            """.format(self.v1,self.v4,self.v5,self.v6,self.v7)
            print(txt)
        
        
    def initiate_homework(self):
    # Parameters:
    # stdID is an integer used to seed the rng
        # IF stdID is given the string "test", it's supposed to return the string "test"
        # This string is then used to thest functions using standard values
    # counter is an integer used to define how many rngs are created.
        #print("Matriculation-Number accepted.\n\nGenerating random numbers..")
        #print("Matriculation-Number accepted.\n\nGenerating personalized homework parameters..")
        if self.stdID == "testing":
            print("#####\n\nTesting functionality\n\n#####")
            print("#####\n\The Std-ID is not a valid Student-ID\n and CAN'T generate a valid code!\n\n#####")
            # NOTE: no self.listofrndms is created in test mode
        else:
        #RNG for this homework
        #IMPORTANT: rd.seed() is set GLOBALLY, as far as we observed
        #this means: the solution-codes are ALWAYS IDENTICAL FOR A UNIQUE STid
        #if rd.seed() is not refreshed after generation of listofrndms.
        #In this version, the generated solution-codes are supposed to be random, therefore. rd.seed() gets refreshed
            rd.seed(self.stdID)
            self.listofrndms = [rd.random() for i in range(self.numberofrandomnumbers)] #create new random numbers that can be used to modify parameters
            rd.seed()
        return
    
    def create_question_variables(self):
        #function returns all generated homework parameters and a boolean 
        #thats used to toggls help_question_1_toggle to  0 (helpoption was not used)
        if self.stdID != "testing":
            #print("\nGenerating personalized homework parameters...\n")
            
            ###PARAMATERS for QUESTION 2           
            ### STEP 1: Create modified parameters
            self.v1 = 19+round(self.listofrndms[0]*2,1) #range 19.0-21.0
            self.v4 = 30+round(self.listofrndms[1]*2,1) #range 30.0-32.0
            self.v5 = 9+round(self.listofrndms[2]*1,1) #range 9.0-10.0
            self.v6 = 4.5+round(self.listofrndms[3]*1,1) #range 4.5.0-5.5
            self.v7 = 4.5+round(self.listofrndms[4]*1,1) #range 4.5.0-5.5 
            
            ### STEP 2: Determine if the Equations are consistent or not:
            ### MOVED TO QUESTION-2-SOLVER
                
        
        elif self.stdID == "testing":
            ###PARAMATERS for QUESTION 2
            self.v1 = 19.3
            self.v4 = 31.1
            self.v5 = 9.5
            self.v6 = 5.5
            self.v7 = 5.0
            
        return
            
###############
###############
###############
    def test_homework_five(self, litmatr, test):
        testnumber = "123456789"
        #Parameters:
        #listofbalances: Student solutions. Will be comparied with opimal solution and decides if task was done correctly or not
        #studentindep = Student solution part 2
        
        #NO parameters were changed for this question
        #hq1tgl: boolean. Decides if the student used the help-option or not

        #output formula:
        # X-YZ
        # X: was the help used? [yes] (randrange(5,9) or [no] randrange(0,4)
        #    -> randomint
        # YZ: Summ(studentID[0]+studentID[1]) [<10] randrange(0,4)) OR [>= 10] randrange(5,9)):
        #    -> 2x randomint
        if test != testnumber:
            return
        
        print("\n####\nQuestion 1\n#####\n")
        
        if litmatr == None:
        
            litmatr = [[1,-1,0,0,0,0,0,0],
            [-1,2,-1,0,0,0,0,0],
            [1,0,1,-1,-1,0,0,0],
            [0,0,0,0,1,-1,-1,0],
            [0,1,1,0,0,1,0,-1],
            [0,2,0,-1,0,0,-2,0]]
        
        #Calculate the independents of litmatr
        _, litinds = sympy.Matrix(litmatr).T.rref()
        print("Number of independent reactions in matrix:",len(litinds))
        print("Position of independent reactions in matrix:",litinds)
        
        print("\n####\nQuestion 2\n#####\n")
        

        concounter = 0
        noncon = []

        # background check relations:
        # v1 = v2 = v3 = (0.5v7 + 0.5v8)
        # 0.5v5 = v6 = v7

        # v5 - v6 - v7 = 0 --> 2v6 - v6 - v7 = 0 --> v6 - v7 = 0
        # v5 - v6 - v7 = 0 --> v5 - 2v6/2v7 = 0
        # 2v2 - v4 - 2v6 --> 2v1 - v4 -2v6/2v7 = 0

        # => 2v1 - v4 = 2v6/2v7 = v5 with v6 - v7 = 0
        p1 = [2*self.v1 - self.v4, 2*round(self.v1*0.05,2) + round(self.v4*0.05,2),"v1, v4"]
        p2 = [self.v5, round(self.v5*0.05,2),"v5"]
        p3 = [2*self.v6, round(2*self.v6*0.05,2),"v6"]
        p4 = [2*self.v7, 2*round(self.v7*0.05,2),"v7"]

        # Check 1
        if (self.v6 - self.v7) + (self.v6*0.05 + self.v7*0.05) >= 0 > (self.v6 - self.v7) - (self.v6*0.05 + self.v7*0.05) or (self.v7 - self.v6) + (self.v6*0.05 + self.v7*0.05) >= 0 > (self.v7 - self.v6) - (self.v6*0.05 + self.v7*0.05):
            concounter += 0            
        else:
            concounter += 1
            noncon.append([p3,p4])

        #Check 2 - 4
        if (p1[0]-p2[0])+(p1[1]+p2[1]) >= 0 > (p1[0]-p2[0])-(p1[1]+p2[1]) or (p2[0]-p1[0])+(p2[1]+p1[1]) >= 0 > (p2[0]-p1[0])-(p2[1]+p1[1]):
            concounter += 0
        else:
            concounter += 1
            noncon.append([p1,p2])
        if (p1[0]-p3[0])+(p1[1]+p3[1]) >= 0 > (p1[0]-p3[0])-(p1[1]+p3[1]) or (p3[0]-p1[0])+(p3[1]+p1[1]) >= 0 > (p3[0]-p1[0])-(p3[1]+p1[1]):
            concounter += 0
        else:
            concounter += 1
            noncon.append([p1,p3])
        if (p2[0]-p3[0])+(p2[1]+p3[1]) >= 0 > (p2[0]-p3[0])-(p2[1]+p3[1]) or (p3[0]-p2[0])+(p3[1]+p2[1]) >= 0 > (p3[0]-p2[0])-(p3[1]+p2[1]):
            concounter += 0
        else:
            concounter += 1
            noncon.append([p2,p3])
        if (p1[0]-p4[0])+(p1[1]+p4[1]) >= 0 > (p1[0]-p4[0])-(p1[1]+p4[1]) or (p4[0]-p1[0])+(p4[1]+p1[1]) >= 0 > (p4[0]-p1[0])-(p4[1]+p1[1]):
            concounter += 0
        else:
            concounter += 1
            noncon.append([p1,p4])
        if (p2[0]-p4[0])+(p2[1]+p4[1]) >= 0 > (p2[0]-p4[0])-(p2[1]+p4[1]) or (p4[0]-p2[0])+(p4[1]+p2[1]) >= 0 > (p4[0]-p2[0])-(p4[1]+p2[1]):
            concounter += 0
        else:
            concounter += 1
            noncon.append([p2,p4])

    #Calculation of the solution:
        print("Concounter=",concounter)
        print("Glucose/v1",self.v1)
        print("Lactate/v4",self.v4)
        print("v5",self.v5)
        print("v6",self.v6)
        print("v7",self.v7)
        print("Nonconform pairs:",noncon)
        print("2v1-v4-v5 = 0 | {} +- {}".format((p1[0]-p2[0]),(p1[1]+p2[1])))
        print("2v1-v4-2v6 = 0 | {} +- {}".format((p1[0]-p3[0]),(p1[1]+p3[1])))
        print("v5-2v6 = 0 | {} +- {}".format((p2[0]-p3[0]),(p2[1]+p3[1])))
        print("2v1-v4-2v7 = 0 | {} +- {}".format((p1[0]-p4[0]),(p1[1]+p4[1])))
        print("v5-2v7 = 0 | {} +- {}".format((p2[0]-p4[0]),(p2[1]+p4[1])))
        print("v6-v7 = 0 | {} +- {}".format((self.v6 - self.v7),(self.v6*0.05 + self.v7*0.05)))    
        #return

             
    ######
    # functions that create widgets
    #
    #
    #
    ###
    
    def build_init_display(self):
        StudID = widgets.IntText( description='Student ID:', value=0 )
        button_StudID   = widgets.Button( description='Generate Parameters' )
        output = widgets.Output()

        disp=widgets.VBox([StudID,button_StudID, output]) 
        return disp
    
###############
###############
###############
    def build_question_display(self):
            listofwidgets = []

            # build question label/output (top 0)
            # build 'Get help button' (top 1)
            # build output (for commentsof help button) (top 2)
            listofwidgets.append(widgets.Output())

            listofwidgets.append(widgets.Button(
                value=False,
                description='Calculation Help',
                disabled=False,
                button_style='', # 'success', 'info', 'warning', 'danger' or ''
                tooltip='Click to get a tipp',
            #     icon='check' # (FontAwesome names without the `fa-` prefix)
            ))

            listofwidgets.append(widgets.Output())

            disp = widgets.VBox(listofwidgets)
            return disp
    
    ######
    # ON-CLICK functions
    # 2 per question
    #
    #
    ###
    def on_init_homework_button_clicked(self, click):
        self.init_display.children[1].disabled = True
        StdID = self.init_display.children[0].value
        self.start_homework(StdID)
    
    def on_question_one_help_button_clicked(self, click):
        self.hq1tgl = 1
        self.question_one_display.children[1].disabled = True
        with self.question_one_display.children[-1]:
            txt = """
##########\n
The task is to establish the independent balances of the stoichiometric matrix G:\n\n
* First step is to establish all balances for the six intracellular metabolites.\nThis is also described in the scripts 'Stoichiometric models I' p. 18.\n\n
* Second step is the gaussian elimination of the converted matrix\nThis is also described in the scripts 'Stoichiometric models I' p. 20-21.\n\n
The number of independent balances can be read off from there.\n
##########\n
            """#HHAVE TO HAVE ANOTHE RLOOK AT THIS
            print(txt)
        
    def on_question_two_help_button_clicked(self, click):
        self.hq2tgl = 1
        self.question_two_display.children[1].disabled = True
        with self.question_two_display.children[-1]:
            txt = """
##########\n
This task is to determine if the observed flux values are consistent or not.\nFor this scenario, consistency is proven when all established dependent relationships\nare consistent with the observed metabolic flux-rates\n\n
* Calculate the standard deviation for all experimental flux rates
* In Question 1, you have calculated the interdepencies of the flux equations of the stoichiometric matrix\nuse these established relationships between fluxes to determine\nif they are consistent to each other.
* If you find a pair of dependent fluxes that are not consistent to one another, the system itself can’t be consistent.\n
Example:\nv_Acetate - v_Ethanol = 0 |  (5.0 - 5.5) +/- (0.25 + 0.275) = 0\n 
##########\n
            """#HHAVE TO HAVE ANOTHE RLOOK AT THIS
            print(txt)
            
##########
##########
##########
#class for homework 7
##########
##########
##########
class create_question_homework_seven():
    #Innit: a question is comprised of the actual
    #Question (string)
    #and a list of answer_options (string) and an int if they are true (1) or not (0)
    def __init__(self, question = str, answers = [], image = None):
        if image != None:
            try:
                file = open(image, "rb")
                image = file.read()
                self.image_widget = self.create_image_widget(image)
            except:
                print("No image to load!")
                self.image_widget = None
        else:
            self.image_widget = None
        self.question = question
        self.answerobjects = []
        self.create_answer_option_objects(answers)
        #self.answerobjects = self.create_answer_option_objects()
        #self.answerobjects2 = []
        
    #create a label with the question
    def create_question_label(self):
        self.question_label  = widgets.Label(value=self.question)
        return
    
    def create__verify_button(self):
        self.verify_button = widgets.Button(
            description='Verify answer',
            disabled=False,
            button_style='warning', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Click here to verify your answer!',
            icon='check' # (FontAwesome names without the `fa-` prefix)
        )
        return
    
    def create_image_widget(self, image):
        i =  widgets.Image(
            value=image,
            format='png',
            width=300,
            height=300,
        )
        return i
    
    #def create_answer_option_objects(self):
    #    list_of_objects = []
    #    for i in self.answers:
    #        y = create_answer_option(i[0],i[1])
    #        list_of_objects.append(y)
    #    return list_of_objects
    
    def create_answer_option_objects(self, original_list_of_answers):
        list_of_answers = copy.deepcopy(original_list_of_answers)
        x = len(list_of_answers)
        if x == 1:
            self.answerobjects.append(create_answer_option(list_of_answers[0]))
            return
        if x == 0:
            print('Something went wrong')
            return
        else:
            y = rd.randint(0,x-1)
            self.answerobjects.append(create_answer_option(list_of_answers[y]))
            del list_of_answers[y]
            self.create_answer_option_objects(list_of_answers)
            return
    
    def check_answer_option_objects(self):
        for i in self.answerobjects:
            i.check_anwer_option()
        return
    
    def get_full_question(self):
        l = []
        self.create_question_label()
        self.create__verify_button()
        l.append(self.question_label)
        
        if self.image_widget != None:
            l.append(self.image_widget)
        
        for i in self.answerobjects:
            l.append(i.button)
        
        l.append(self.verify_button)
        wbox = widgets.VBox(l)
        return wbox
    
    def on_button_clicked(self, click):
        self.verify_button.disabled = True
        self.check_answer_option_objects()

class create_answer_option():
    #this takes the list of answer options of a question and turns it into toogle.button widgets
    #Position 0 in the list is the descripion
    #Position 1 in the list is if the answer option is True(1) or false (0)
    #Position 2 in the list is for the case that it's not a True-False question.
        #this is why len(inputlist) is checked
        
    #Note: this is backwards and should be exchanged through a *args but the def calling these
    # create_answer_option objects needs to call the positions in the greater list/dict, so that isnt possible (YET)
    def __init__(self, inputlist):
        self.descr = inputlist[0]
        self.yesorno = inputlist[1]
        
        if len(inputlist) == 3:
            optionlist = inputlist[2]
        else:
            optionlist = [("True",1),("False",0)]
        
        self.optionlist = optionlist
        self.button = self.create_answer_option_button()

    #create a toggle button for a single answer_option
    def create_answer_option_button(self):
        w = widgets.ToggleButtons(
            options= self.optionlist,
            description=self.descr,
            disabled=False,
            button_style='', # 'success', 'info', 'warning', 'danger' or ''
           # tooltips=['Description of slow', 'Description of regular', 'Description of fast'],
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='auto', height='40px')
        )
        return w
    
    def check_anwer_option(self):
        if self.button.value == self.yesorno:
            if self.button.value == 1:
                self.button.icons = ["check",""]
            if self.button.value == 0:
                self.button.icons = ["","check"]
            self.button.button_style = 'success'
            self.button.disabled = True
        elif self.button.value != self.yesorno:
            if self.button.value == 1:
                self.button.icons = ["","check"]
            if self.button.value == 0:
                self.button.icons = ["check",""]
            self.button.button_style = "danger"
            self.button.disabled = True

##########
##########
##########            
class create_homework_seven_task():
    def __init__(self, question, solution):
        self.question = question
        self.sol = solution
        self.create_int_text()
        self.create_question_label()
        
    def create_question_label(self):
        self.question_label = widgets.Label(
            value=self.question,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='auto')    
        )
    
    def create_int_text(self):
        self.int_text = widgets.IntText(
            value=0,
            description='Answer:',
            tooltip ='Answer:',
            disabled=False,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='auto', height='40px')
        )
    
    def create_verify_button(self):
        self.verify_button = widgets.Button(
            description='Verify answer',
            disabled=False,
            button_style='warning', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Click here to verify your answer!',
            icon='check', # (FontAwesome names without the `fa-` prefix),
            display='flex',
            align_items='center',
            flex_flow='column',
            #align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='50%', height='40px')
        )
        
    def on_verify_button_clicked(self, click):
        if self.int_text.value == self.sol:
            self.verify_button.disabled = True
            self.int_text.disabled = True
            self.verify_button.button_style = 'success'
            self.verify_button.description="Correct answer!"
        else:
            #self.verify_button.disabled = True
            #self.int_text.disabled = True
            self.verify_button.button_style = 'danger'
            self.verify_button.description="Wrong answer. Please try again!"
            
    def get_full_question(self):
        self.create_verify_button()
        #self.create_help_button()
        #i = [self.verify_button, self.help_button]
        #hbox = widgets.HBox(i)
        
        j = [self.question_label, self.int_text,self.verify_button]
        wbox = widgets.VBox(j)
        return wbox
##########
##########
##########
class create_homework_seven_task_one_one():
    def __init__(self, matricnb = int):
        self.init_display = self.build_init_display()
        self.question =  "What is the maximum specific growth rate? [1/h]"
        #self.create_solution() #needs self.ID from self.init_display #gets moved to on-click event in init-display
        self.create_float_text()
        self.create_question_label()
        self.output = widgets.Output()
        #self.ID created by on-click event on button in self.init_display
        #self.sol created by  self.create_solution()
    
    def create_solution(self, ID):
        if type(ID) == int:
            rd.seed(ID)
            listofrndms = [rd.random() for i in range(10)]
            #self.sol = 0.300+listofrndms[9]*0.100
            self.sol = 0.386
        elif matricnb == "Test":
            self.sol = 0.386
        else:
            with self.init_display.children[-1]:
                print("matriculation number has to be an integer!")
                #self.output.clear_output(wait=True)
    
    def create_question_label(self):
        self.question_label = widgets.Label(
            value=self.question,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='auto')    
        )
    
    def create_float_text(self):
        # create two Float-TExt widgets that get implemented into the question interface
        # self.float_corr is the correlation coefficient R
        # self.float_mue is the growth rate µ
        # both user-inputs can be checked by taking the value in 'widget'.value
        self.float_corr = widgets.FloatText(
            value=0,
            description='Correlation Coefficient\n|| Example: R^2 = 0.00 (Float)',
            tooltip ='Correlation Coefficient\n|| Example: R^2 = 0.00 (Float)',
            disabled=False,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='auto', height='40px')
        )
        self.float_mue = widgets.FloatText(
            value=0,
            description='maximum specific growthrate\n|| Example: µ = 1.00 (Float)',
            tooltip ='maximum specific growthrate\n|| Example: µ = 1.00 (Float)',
            disabled=False,
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='auto', height='40px')
        )
    
    def create_verify_button(self):
        self.verify_button = widgets.Button(
            description='Verify answer',
            disabled=False,
            button_style='warning', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Click here to verify your answer!',
            icon='check', # (FontAwesome names without the `fa-` prefix),
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='auto', height='40px')
        )
            
    def create_help_button(self):
        self.help_button = widgets.Button(
            description='Help required?',
            disabled=False,
            button_style='info', # 'success', 'info', 'warning', 'danger' or ''
            tooltip='Click here to get a tipp!',
            display='flex',
            flex_flow='column',
            align_items='stretch',
            style= {'description_width': 'initial'},
            layout = widgets.Layout(width='auto', height='40px')
        )
        
    def on_verify_button_clicked(self, click):
        self.verify_button.button_style = 'warning'
        self.verify_button.description="Wrong answer. Please try again!"
        #First step: Check if correlation coefficient is > 0.95
        #If yes: move on to growth-rate
        #if no: depict on verify button and return
        if self.float_corr.value > 1.00:
            self.verify_button.button_style = 'warning'
            self.verify_button.description="Correlation Coefficient can't be > 1.00. Please try again!"
        elif 0.95 <= self.float_corr.value <= 1.00:
        #Second Step: check if growth rate solution is correct (is equal to self.sol) 
        #If yes: answer correct
        #If no: Answer wrong
        #If close to correc: depict on verify button
            if round(self.float_mue.value,2) == round(self.sol,2):
                self.verify_button.disabled = True
                self.float_corr.disabled = True
                self.float_mue.disabled = True
                self.verify_button.button_style = 'success'
                self.verify_button.description="Correct answer!"
                return
            elif round(self.float_mue.value,1) == round(self.sol,1):
                self.verify_button.button_style = 'danger'
                self.verify_button.description="Wrong answer, but close! Please try again!"
                return
        else:
            self.verify_button.button_style = 'warning'
            self.verify_button.description="Correlation Coefficient has to be > 0.95. Please try again!"
                
            
    def on_help_button_clicked(self, click):
        self.help_button.disabled = True
        txt = """
##########\n
For cells growing at a constant specific growth rate µ, the biomass concentration x increases with time:\n
x_i = x_0 * e^m\n
µ can be determined by plotting the natural logarithm of the OD against the time h\nand determining the slop of the exponential growth phase\n 
µ = (ln(x_o) - ln(x_i)) / t\n\n
Note:\n
* To assume a linear correlation between µ and t, atleast three data points are required\n
* The correlation coefficient R^2 has to be > 0.95 
##########\n
        """
        with self.output:
            print(txt)
            
    #on-click in init-display:
    #store Student-ID in self.ID
    def on_init_homework_button_clicked(self, click):
        self.init_display.children[1].disabled = True
        ID = self.init_display.children[0].value
        self.create_solution(ID)
        with self.init_display.children[2]:
            print("#####\nStudent-ID accepted!\nGood luck!\n#####")

    
    def get_full_question(self):
        self.create_verify_button()
        self.create_help_button()
        i = [self.verify_button, self.help_button]
        hbox = widgets.HBox(i)
        
        j = [self.question_label, self.float_corr, self.float_mue, hbox, self.output]
        wbox = widgets.VBox(j)
        return wbox
    
    def build_init_display(self):
        StudID = widgets.IntText( description='Student ID:', value=0 )
        button_StudID   = widgets.Button( description='Generate Parameters' )
        output = widgets.Output()

        disp=widgets.VBox([StudID,button_StudID, output]) 
        return disp

###
def full_question_objects():
    q_1_1_answers = [
        ["for butanol between 17 h - 32 h from the start of the process",0],
        ["for neither butanol or butyrate",0],
        ["for butyrate between 8 h - 14 h from the start of the process",1]    
    ]
    q_1_2_answers = [
        ["is reached after complete glucose degradiation",0],
        ["is 18.4 mM/h",1],
        ["is reached before complete butyrate degradiation",1],
        ["is 13.3 mM/h",0]
    ]
    q_1_3_1_answers = [
        ["increases linearly",0],
        ["decreeases linearly",0],
        ["increases exponentially",1],
        ["decreases exponentially",0],
        ["stays constant",0]
    ]
    q_1_3_2_answers = [
        ["stays the same",0],
        ["exponentially",1],
        ["linearly",0]
    ]
    q_1_3_3_answers = [
        ["a linear increase in product concentration",0],
        ["an exponential increase in product concentration",1]
    ]
    q_1_3_4_answers = [
        ["the kinetic orders of both formation processes are the same",1],
        ["the formation rates of both formation processes are the same",0],
        ["product formation is growth-dependent",1],
        ["product formation is growth-independent",0]
    ]
    q_2_1_answers = [
        ["1+2 are organic solvents",0],
        ["3+4 are organic solvents",1],
        ["1+2 are organic acids",1],
        ["3+4 are organic acids",0]
    ]
    q_2_2_answers = [
        ["Acetyl-CoA ↔ Acetat + ATP\nAcetoacetyl-CoA + Butyrat → Aceton + CO2 + Butyryl-CoA",0],
        ["Butyryl-CoA ↔ Butyrat + ATP\nAcetoacetyl-CoA + Acetat → Aceton + CO2 + Acetyl-CoA",0],
        ["Butyryl-CoA ↔ Butyrat + ATP\nAcetoacetyl-CoA + Butyrat → Aceton + CO2 + Butyryl-CoA",1],
        ["Acetate_extracellular ↔ Acetate_intracellular\nAcetoacetyl-CoA + Acetat → Aceton + CO2 + Acetyl-CoA",1],
        ["Acetyl-CoA ↔ Acetat + ATP\nAcetoacetyl-CoA + Acetat → Aceton + CO2 + Acetyl-CoA",1]
    ]
    q_2_3_1_answers = [
        ["not be determined by determining the concentration over time using HPLC,\nsince the metabolic network has too many degrees of freedom (is underdetermined).",1],
        ["be determined by determining the concentration over time using HPLC.",0],
        ["not be determined by determining concentration over time using HPLC since acetone is volatile.",0]
    ]
    q_2_3_2_answers = [
        ["undetermined; not all intracellular rates can be calculated.",1],
        ["over-determined; all intracellular rates can be calculated and error estimation performed.",0],
        ["determined; all intracellular rates can be determined.",0]
    ]
    q_3_1_answers = [
        ["Acetyl-CoA ↔ Acetat + ATP\nAcetoacetyl-CoA + Acetat → Aceton + CO2 + Acetyl-CoA",0],
        ["Butyryl-CoA ↔ Butyrat + ATP",1],
        ["Acetoacetyl-CoA → Acceton + CO2 + ATP",1],#added Aceton on Product side (Quant-MiBi_Calc-Guide.pdf, p. 30)
        ["Acetyl-P ↔ Acetat + ATP",0],
        ["Butyryl-CoA ↔ Butyrat + ATP\nAcetoacetyl-CoA + Butyrat → Aceton + CO2 + Butyryl-CoA",0],
        ["Acetyl-CoA ↔ Acetat + ATP",1]
    ]
    q_3_2_answers = [
        ["",0,[("Yes",1),("No",0)]]
    ]
    question_dict = {}
    question_dict["q_1_1"] = ["The formation is directly proportional to the biomass increase:",q_1_1_answers, None]
    question_dict["q_1_2"] = ["The maximum rate of butanol formation:",q_1_2_answers, None]
    question_dict["q_1_3_1"] = ["If the biomass concentration increases exponentially over time, the biomass formation rate is:",q_1_3_1_answers, None]
    question_dict["q_1_3_2"] = ["For cells that proliferate at a constant specific growth rate µ, the concentration of biomass increases over time:",q_1_3_2_answers, None]
    question_dict["q_1_3_3"] = ["Product formation according to first-order kinetics corresponds to:",q_1_3_3_answers, None]
    question_dict["q_1_3_4"] = ["If there is a linear relationship between the product and biomass concentration r_x:",q_1_3_4_answers, None]
    question_dict["q_2_1"] = ["Which C. acetobutylicum products are represented by curves 1, 2, 3, and 4?",q_2_1_answers, "data/801059_Graph_Blank1997_Prozess_C_acetobutylicum_lora_080703_v3_modified4question.png"]
    question_dict["q_2_2"] = ["Which reaction equations represent a possible synthetic route to acetone?",q_2_2_answers, None]
    question_dict["q_2_3_1"] = ["The reaction rates to acetone can:",q_2_3_1_answers, None]
    question_dict["q_2_3_2"] = ["The metabolic (stoichiometric) model is:",q_2_3_2_answers, None]
    question_dict["q_3_1"] = ["Which reaction equations represent the net equations for butyrate, acetate and acetone?",q_3_1_answers, None]
    question_dict["q_3_2"] = ["Is the acetone synthesis rate determinable?",q_3_2_answers, None]
    
    
    list_of_objects = []
    for i in question_dict.keys():
        list_of_objects.append(create_question_homework_seven(question_dict[i][0],question_dict[i][1], question_dict[i][2]))
    return list_of_objects

def full_task_objects():
    #t_1_1 = [
    #    "What is the maximum specific growth rate? [1/h]",0.368
    #]
    t_2_1_1 = [
        "How many reaction rates does the stoichiometric model representing the metabolic network of C. acetobutylicum contain?",20
    ]
    t_2_1_2 = [
     "How many reaction pathways are there for the synthesis of acetone in C. acetobutylicum?",2   
    ]
    t_2_2_1 = [
     "How many rates are in r_m?",8
    ]
    t_2_2_2 = [
     "The number of rates r_c to calculate is:",12#is 12 correct?   
    ]
    t_2_2_3 = [
     "How many of (all) the rates can be determined using liquid chromatography (HPLC with any detector) and a spectrophotometer?",8   
    ]
    t_3_1 = [
     "How many reaction equations does the simplified metabolic model for calculating the intracellular reaction rates including acetone production contain?",11   
    ]

    task_dict = {}
    #task_dict["t_1_1"] = t_1_1 #NOTE: Task 1.1 requires calculation and a float value as answer. because of this, it was kept seperate.
    task_dict["t_2_1_1"] = t_2_1_1
    task_dict["t_2_1_2"] = t_2_1_2
    task_dict["t_2_2_1"] = t_2_2_1
    task_dict["t_2_2_2"] = t_2_2_2
    task_dict["t_2_2_3"] = t_2_2_3
    task_dict["t_3_1"] = t_3_1
    
    
    list_of_objects = []
    for i in task_dict.keys():
        list_of_objects.append(create_homework_seven_task(task_dict[i][0],task_dict[i][1]))
    return list_of_objects